from typing import Set, Dict, List, Tuple, Any, Optional, TYPE_CHECKING
from collections import defaultdict
import logging

from ailment import AILBlockWalker
from ailment.block import Block
from ailment.statement import Statement, Assignment, Store, Call, ConditionalJump
from ailment.expression import (
    Register,
    Convert,
    Load,
    StackBaseOffset,
    Expression,
    DirtyExpression,
    VEXCCallExpression,
    Tmp,
    Const,
    BinaryOp,
)

from ...engines.light import SpOffset
from ...code_location import CodeLocation, ExternalCodeLocation
from ...sim_variable import SimStackVariable, SimMemoryVariable
from ...knowledge_plugins.propagations.states import Equivalence
from ...knowledge_plugins.key_definitions import atoms
from ...knowledge_plugins.key_definitions.definition import Definition
from ...knowledge_plugins.key_definitions.constants import OP_BEFORE
from .. import Analysis, AnalysesHub
from .ailgraph_walker import AILGraphWalker
from .expression_narrower import ExpressionNarrowingWalker
from .block_simplifier import BlockSimplifier
from .ccall_rewriters import CCALL_REWRITERS
from .expression_counters import SingleExpressionCounter

if TYPE_CHECKING:
    from ailment.manager import Manager
    from angr.analyses.reaching_definitions import ReachingDefinitionsModel


_l = logging.getLogger(__name__)


class HasCallNotification(Exception):
    """
    Notifies the existence of a call statement.
    """


class AILBlockTempCollector(AILBlockWalker):
    """
    Collects any temporaries used in a block.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.temps = set()
        self.expr_handlers[Tmp] = self._handle_Tmp

    # pylint:disable=unused-argument
    def _handle_Tmp(self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement, block) -> None:
        if isinstance(expr, Tmp):
            self.temps.add(expr)


class AILSimplifier(Analysis):
    """
    Perform function-level simplifications.
    """

    def __init__(
        self,
        func,
        func_graph=None,
        remove_dead_memdefs=False,
        stack_arg_offsets: Optional[Set[Tuple[int, int]]] = None,
        unify_variables=False,
        ail_manager: Optional["Manager"] = None,
        gp: Optional[int] = None,
        narrow_expressions=False,
        only_consts=False,
        fold_callexprs_into_conditions=False,
        use_callee_saved_regs_at_return=True,
    ):
        self.func = func
        self.func_graph = func_graph if func_graph is not None else func.graph
        self._reaching_definitions: Optional["ReachingDefinitionsModel"] = None
        self._propagator = None

        self._remove_dead_memdefs = remove_dead_memdefs
        self._stack_arg_offsets = stack_arg_offsets
        self._unify_vars = unify_variables
        self._ail_manager = ail_manager
        self._gp = gp
        self._narrow_expressions = narrow_expressions
        self._only_consts = only_consts
        self._fold_callexprs_into_conditions = fold_callexprs_into_conditions
        self._use_callee_saved_regs_at_return = use_callee_saved_regs_at_return

        self._calls_to_remove: Set[CodeLocation] = set()
        self._assignments_to_remove: Set[CodeLocation] = set()
        self.blocks = {}  # Mapping nodes to simplified blocks

        self.simplified: bool = False
        self._simplify()

    def _simplify(self):
        if self._narrow_expressions:
            _l.debug("Narrowing expressions")
            narrowed_exprs = self._narrow_exprs()
            self.simplified |= narrowed_exprs
            if narrowed_exprs:
                _l.debug("... expressions narrowed")
                self._rebuild_func_graph()
                self._clear_cache()

        _l.debug("Folding expressions")
        folded_exprs = self._fold_exprs()
        self.simplified |= folded_exprs
        if folded_exprs:
            _l.debug("... expressions folded")
            self._rebuild_func_graph()
            # reaching definition analysis results are no longer reliable
            self._clear_cache()

        if self._only_consts:
            return

        _l.debug("Rewriting ccalls")
        ccalls_rewritten = self._rewrite_ccalls()
        self.simplified |= ccalls_rewritten
        if ccalls_rewritten:
            _l.debug("... ccalls rewritten")
            self._rebuild_func_graph()
            self._clear_cache()

        if self._unify_vars:
            _l.debug("Removing dead assignments")
            r = self._remove_dead_assignments()
            if r:
                _l.debug("... dead assignments removed")
                self.simplified = True
                self._rebuild_func_graph()
                self._clear_cache()

            _l.debug("Unifying local variables")
            r = self._unify_local_variables()
            if r:
                _l.debug("... local variables unified")
                self.simplified = True
                self._rebuild_func_graph()

            # _fold_call_exprs() may set self._calls_to_remove, which will be honored in _remove_dead_assignments()
            _l.debug("Folding call expressions")
            r = self._fold_call_exprs()
            if r:
                _l.debug("... call expressions folded")
                self.simplified = True
                self._rebuild_func_graph()

        _l.debug("Removing dead assignments")
        r = self._remove_dead_assignments()
        if r:
            _l.debug("... dead assignments removed")
            self.simplified = True
            self._rebuild_func_graph()

    def _rebuild_func_graph(self):
        def _handler(node):
            return self.blocks.get(node, None)

        AILGraphWalker(self.func_graph, _handler, replace_nodes=True).walk()
        self.blocks = {}

    def _compute_reaching_definitions(self) -> "ReachingDefinitionsModel":
        # Computing reaching definitions or return the cached one
        if self._reaching_definitions is not None:
            return self._reaching_definitions
        rd = self.project.analyses.ReachingDefinitions(
            subject=self.func,
            func_graph=self.func_graph,
            # init_context=(),    <-- in case of fire break glass
            observe_all=False,
            use_callee_saved_regs_at_return=self._use_callee_saved_regs_at_return,
            track_tmps=True,
        ).model
        self._reaching_definitions = rd
        return rd

    def _compute_propagation(self, immediate_stmt_removal: bool = False):
        # Propagate expressions or return the existing result
        if self._propagator is not None:
            return self._propagator
        prop = self.project.analyses.Propagator(
            func=self.func,
            func_graph=self.func_graph,
            gp=self._gp,
            only_consts=self._only_consts,
            reaching_definitions=self._compute_reaching_definitions(),
            immediate_stmt_removal=immediate_stmt_removal,
        )
        self._propagator = prop
        return prop

    def _clear_cache(self) -> None:
        self._propagator = None
        self._reaching_definitions = None

    def _clear_propagator_cache(self) -> None:
        self._propagator = None

    def _clear_reaching_definitions_cache(self) -> None:
        self._reaching_definitions = None

    #
    # Expression narrowing
    #

    def _narrow_exprs(self) -> bool:
        """
        A register may be used with full width even when only the lower bytes are really needed. This results in the
        incorrect determination of wider variables while the actual variable is narrower (e.g., int64 vs char). This
        optimization narrows a register definition if all its uses are narrower than the definition itself.
        """

        narrowed = False

        addr_and_idx_to_block: Dict[Tuple[int, int], Block] = {}
        for block in self.func_graph.nodes():
            addr_and_idx_to_block[(block.addr, block.idx)] = block

        rd = self._compute_reaching_definitions()
        sorted_defs = sorted(rd.all_definitions, key=lambda d: d.codeloc, reverse=True)
        for def_ in (d_ for d_ in sorted_defs if d_.codeloc.context is None):
            if isinstance(def_.atom, atoms.Register):
                needs_narrowing, to_size, use_exprs = self._narrowing_needed(def_, rd, addr_and_idx_to_block)
                if needs_narrowing:
                    # replace the definition
                    if not isinstance(def_.codeloc, ExternalCodeLocation):
                        old_block = addr_and_idx_to_block.get((def_.codeloc.block_addr, def_.codeloc.block_idx))
                        if old_block is None:
                            # this definition might be inside a callee function, which is why the block does not exist
                            # ignore it
                            continue

                        the_block = self.blocks.get(old_block, old_block)
                        stmt = the_block.statements[def_.codeloc.stmt_idx]
                        r, new_block = False, None
                        if isinstance(stmt, Assignment) and isinstance(stmt.dst, Register):
                            tags = dict(stmt.dst.tags)
                            tags["reg_name"] = self.project.arch.translate_register_name(
                                def_.atom.reg_offset, size=to_size
                            )
                            tags["write_size"] = stmt.dst.size
                            new_assignment_dst = Register(
                                stmt.dst.idx,
                                None,
                                def_.atom.reg_offset,
                                to_size * self.project.arch.byte_width,
                                **tags,
                            )
                            new_assignment_src = Convert(
                                stmt.src.idx,  # FIXME: This is a hack
                                stmt.src.bits,
                                to_size * self.project.arch.byte_width,
                                False,
                                stmt.src,
                                **stmt.src.tags,
                            )
                            r, new_block = BlockSimplifier._replace_and_build(
                                the_block,
                                {
                                    def_.codeloc: {
                                        stmt.dst: new_assignment_dst,
                                        stmt.src: new_assignment_src,
                                    }
                                },
                                replace_assignment_dsts=True,
                                replace_loads=True,
                            )
                        elif isinstance(stmt, Call):
                            if stmt.ret_expr is not None:
                                tags = dict(stmt.ret_expr.tags)
                                tags["reg_name"] = self.project.arch.translate_register_name(
                                    def_.atom.reg_offset, size=to_size
                                )
                                new_retexpr = Register(
                                    stmt.ret_expr.idx,
                                    None,
                                    def_.atom.reg_offset,
                                    to_size * self.project.arch.byte_width,
                                    **tags,
                                )
                                r, new_block = BlockSimplifier._replace_and_build(
                                    the_block, {def_.codeloc: {stmt.ret_expr: new_retexpr}}
                                )
                        if not r:
                            # couldn't replace the definition...
                            continue
                        self.blocks[old_block] = new_block

                    # replace all uses if necessary
                    for use_loc, (use_type, use_expr_tpl) in use_exprs:
                        if isinstance(use_expr_tpl[0], Register) and to_size == use_expr_tpl[0].size:
                            # don't replace registers to the same registers
                            continue

                        old_block = addr_and_idx_to_block.get((use_loc.block_addr, use_loc.block_idx))
                        the_block = self.blocks.get(old_block, old_block)

                        if use_type in {"expr", "mask", "convert"}:
                            # the first used expr
                            use_expr_0 = use_expr_tpl[0]
                            tags = dict(use_expr_0.tags)
                            tags["reg_name"] = self.project.arch.translate_register_name(
                                def_.atom.reg_offset, size=to_size
                            )
                            new_use_expr_0 = Register(
                                use_expr_0.idx,
                                None,
                                def_.atom.reg_offset,
                                to_size * self.project.arch.byte_width,
                                **tags,
                            )

                            # the second used expr (if it exists)
                            if len(use_expr_tpl) == 2:
                                use_expr_1 = use_expr_tpl[1]
                                assert isinstance(use_expr_1, BinaryOp)
                                con = use_expr_1.operands[1]
                                assert isinstance(con, Const)
                                new_use_expr_1 = BinaryOp(
                                    use_expr_1.idx,
                                    use_expr_1.op,
                                    [
                                        new_use_expr_0,
                                        Const(con.idx, con.variable, con.value, new_use_expr_0.bits, **con.tags),
                                    ],
                                    use_expr_1.signed,
                                    floating_point=use_expr_1.floating_point,
                                    rounding_mode=use_expr_1.rounding_mode,
                                    **use_expr_1.tags,
                                )

                                if use_expr_1.size > new_use_expr_1.size:
                                    new_use_expr_1 = Convert(
                                        None,
                                        new_use_expr_1.bits,
                                        use_expr_1.bits,
                                        False,
                                        new_use_expr_1,
                                        **new_use_expr_1.tags,
                                    )

                                r, new_block = BlockSimplifier._replace_and_build(
                                    the_block, {use_loc: {use_expr_1: new_use_expr_1}}
                                )
                            elif len(use_expr_tpl) == 1:
                                if use_expr_0.size > new_use_expr_0.size:
                                    new_use_expr_0 = Convert(
                                        None,
                                        new_use_expr_0.bits,
                                        use_expr_0.bits,
                                        False,
                                        new_use_expr_0,
                                        **new_use_expr_0.tags,
                                    )

                                r, new_block = BlockSimplifier._replace_and_build(
                                    the_block, {use_loc: {use_expr_0: new_use_expr_0}}
                                )
                            else:
                                _l.warning("Nothing to replace at %s.", use_loc)
                                r = False
                                new_block = None
                        elif use_type == "binop-convert":
                            use_expr_0 = use_expr_tpl[0]
                            tags = dict(use_expr_0.tags)
                            tags["reg_name"] = self.project.arch.translate_register_name(
                                def_.atom.reg_offset, size=to_size
                            )
                            new_use_expr_0 = Register(
                                use_expr_0.idx,
                                None,
                                def_.atom.reg_offset,
                                to_size * self.project.arch.byte_width,
                                **tags,
                            )

                            use_expr_1: BinaryOp = use_expr_tpl[1]
                            # build the new use_expr_1
                            new_use_expr_1_operands = {}
                            if use_expr_1.operands[0] is use_expr_0:
                                new_use_expr_1_operands[0] = new_use_expr_0
                                other_operand = use_expr_1.operands[1]
                            else:
                                new_use_expr_1_operands[1] = new_use_expr_0
                                other_operand = use_expr_1.operands[0]
                            use_expr_2: Convert = use_expr_tpl[2]
                            if other_operand.bits == use_expr_2.from_bits:
                                new_other_operand = Convert(
                                    None, use_expr_2.from_bits, use_expr_2.to_bits, False, other_operand
                                )
                            else:
                                # Some operations, like Sar and Shl, have operands with different sizes
                                new_other_operand = other_operand

                            if 0 in new_use_expr_1_operands:
                                new_use_expr_1_operands[1] = new_other_operand
                            else:
                                new_use_expr_1_operands[0] = new_other_operand

                            # build new use_expr_1
                            new_use_expr_1 = BinaryOp(
                                use_expr_1.idx,
                                use_expr_1.op,
                                [new_use_expr_1_operands[0], new_use_expr_1_operands[1]],
                                use_expr_1.signed,
                                bits=to_size * 8,
                                floating_point=use_expr_1.floating_point,
                                rounding_mode=use_expr_1.rounding_mode,
                                **use_expr_1.tags,
                            )

                            # first remove the old conversion
                            r, new_block = BlockSimplifier._replace_and_build(
                                the_block, {use_loc: {use_expr_2: use_expr_2.operand}}
                            )
                            # then replace use_expr_1
                            if r:
                                r, new_block = BlockSimplifier._replace_and_build(
                                    new_block, {use_loc: {use_expr_1: new_use_expr_1}}
                                )
                        else:
                            raise TypeError(f'Unsupported use_type value "{use_type}"')

                        if not r:
                            _l.warning("Failed to replace use-expr at %s.", use_loc)
                        else:
                            self.blocks[old_block] = new_block

                    narrowed = True

        return narrowed

    def _narrowing_needed(
        self, def_, rd, addr_and_idx_to_block
    ) -> Tuple[bool, Optional[int], Optional[List[Tuple[CodeLocation, Tuple[str, Tuple[Expression, ...]]]]]]:
        def_size = def_.size
        # find its uses
        use_and_exprs = rd.all_uses.get_uses_with_expr(def_)

        all_used_sizes = set()
        used_by: List[Tuple[CodeLocation, Tuple[str, Tuple[Expression, ...]]]] = []

        for loc, expr in use_and_exprs:
            old_block = addr_and_idx_to_block.get((loc.block_addr, loc.block_idx), None)
            if old_block is None:
                # missing a block for whatever reason
                return False, None, None

            block = self.blocks.get(old_block, old_block)
            if loc.stmt_idx >= len(block.statements):
                # missing a statement for whatever reason
                return False, None, None
            stmt = block.statements[loc.stmt_idx]

            expr_size, used_by_exprs = self._extract_expression_effective_size(stmt, expr)
            if expr_size is None:
                # it's probably used in full width
                return False, None, None

            all_used_sizes.add(expr_size)
            used_by.append((loc, used_by_exprs))

        if len(all_used_sizes) == 1 and next(iter(all_used_sizes)) < def_size:
            return True, next(iter(all_used_sizes)), used_by

        return False, None, None

    def _extract_expression_effective_size(
        self, statement, expr
    ) -> Tuple[Optional[int], Optional[Tuple[str, Tuple[Expression, ...]]]]:
        """
        Determine the effective size of an expression when it's used.
        """

        walker = ExpressionNarrowingWalker(expr)
        walker.walk_statement(statement)
        if not walker.operations:
            if expr is None:
                return None, None
            return expr.size, ("expr", (expr,))

        first_op = walker.operations[0]
        if isinstance(first_op, Convert):
            return first_op.to_bits // self.project.arch.byte_width, ("convert", (first_op,))
        if isinstance(first_op, BinaryOp):
            second_op = None
            if len(walker.operations) >= 2:
                second_op = walker.operations[1]
            if (
                first_op.op == "And"
                and isinstance(first_op.operands[1], Const)
                and (second_op is None or isinstance(second_op, BinaryOp) and isinstance(second_op.operands[1], Const))
            ):
                mask = first_op.operands[1].value
                if mask == 0xFF:
                    return 1, ("mask", (first_op, second_op)) if second_op is not None else ("mask", (first_op,))
                if mask == 0xFFFF:
                    return 2, ("mask", (first_op, second_op)) if second_op is not None else ("mask", (first_op,))
                if mask == 0xFFFF_FFFF:
                    return 4, ("mask", (first_op, second_op)) if second_op is not None else ("mask", (first_op,))
            if (
                (first_op.operands[0] is expr or first_op.operands[1] is expr)
                and first_op.op not in {"Shr", "Sar"}
                and isinstance(second_op, Convert)
                and second_op.from_bits == expr.bits
            ):
                return min(expr.bits, second_op.to_bits) // self.project.arch.byte_width, (
                    "binop-convert",
                    (expr, first_op, second_op),
                )

        if expr is None:
            return None, None
        return expr.size, ("expr", (expr,))

    #
    # Expression folding
    #

    def _fold_exprs(self):
        """
        Fold expressions: Fold assigned expressions that are only used once.
        """

        # propagator
        propagator = self._compute_propagation(immediate_stmt_removal=True)
        replacements = propagator.replacements

        # take replacements and rebuild the corresponding blocks
        replacements_by_block_addrs_and_idx = defaultdict(dict)
        for codeloc, reps in replacements.items():
            if reps:
                replacements_by_block_addrs_and_idx[(codeloc.block_addr, codeloc.block_idx)][codeloc] = reps

        if not replacements_by_block_addrs_and_idx:
            return False

        blocks_by_addr_and_idx = {(node.addr, node.idx): node for node in self.func_graph.nodes()}

        if self._stack_arg_offsets:
            insn_addrs_using_stack_args = {ins_addr for ins_addr, _ in self._stack_arg_offsets}
        else:
            insn_addrs_using_stack_args = None

        replaced = False
        for (block_addr, block_idx), reps in replacements_by_block_addrs_and_idx.items():
            block = blocks_by_addr_and_idx[(block_addr, block_idx)]

            # only replace loads if there are stack arguments in this block
            replace_loads = insn_addrs_using_stack_args is not None and {
                stmt.ins_addr for stmt in block.statements
            }.intersection(insn_addrs_using_stack_args)

            r, new_block = BlockSimplifier._replace_and_build(block, reps, gp=self._gp, replace_loads=replace_loads)
            replaced |= r
            self.blocks[block] = new_block

        if replaced:
            # blocks have been rebuilt - expression propagation results are no longer reliable
            self._clear_cache()
        return replaced

    #
    # Unifying local variables
    #

    def _unify_local_variables(self) -> bool:
        """
        Find variables that are definitely equivalent and then eliminate unnecessary copies.
        """

        simplified = False

        prop = self._compute_propagation()
        if not prop.model.equivalence:
            return simplified

        addr_and_idx_to_block: Dict[Tuple[int, int], Block] = {}
        for block in self.func_graph.nodes():
            addr_and_idx_to_block[(block.addr, block.idx)] = block

        equivalences: Dict[Any, Set[Equivalence]] = defaultdict(set)
        atom_by_loc = set()
        for eq in prop.model.equivalence:
            equivalences[eq.atom1].add(eq)
            atom_by_loc.add((eq.codeloc, eq.atom1))

        # sort keys to ensure a reproducible result
        sorted_loc_and_atoms = sorted(atom_by_loc, key=lambda x: x[0])

        for _, atom in sorted_loc_and_atoms:
            eqs = equivalences[atom]
            if len(eqs) > 1:
                continue

            eq = next(iter(eqs))

            # Acceptable equivalence classes:
            #
            # stack variable == register
            # register variable == register
            # stack variable == Conv(register, M->N)
            # global variable == register
            #
            # Equivalence is generally created at assignment sites. Therefore, eq.atom0 is the definition and
            # eq.atom1 is the use.
            the_def = None
            if isinstance(eq.atom0, SimMemoryVariable):  # covers both Stack and Global variables
                if isinstance(eq.atom1, Register):
                    # stack_var == register or global_var == register
                    to_replace = eq.atom1
                    to_replace_is_def = False
                elif isinstance(eq.atom1, Convert) and isinstance(eq.atom1.operand, Register):
                    # stack_var == Conv(register, M->N)
                    to_replace = eq.atom1.operand
                    to_replace_is_def = False
                else:
                    continue

            elif isinstance(eq.atom0, Register):
                if isinstance(eq.atom1, Register):
                    # register == register
                    if self.project.arch.is_artificial_register(eq.atom0.reg_offset, eq.atom0.size):
                        to_replace = eq.atom0
                        to_replace_is_def = True
                    else:
                        to_replace = eq.atom1
                        to_replace_is_def = False
                else:
                    continue

            else:
                continue

            # find the definition of this register
            rd = self._compute_reaching_definitions()
            if to_replace_is_def:
                # find defs
                defs = []
                for def_ in rd.all_definitions:
                    if def_.codeloc == eq.codeloc:
                        if isinstance(to_replace, SimStackVariable):
                            if isinstance(def_.atom, atoms.MemoryLocation) and isinstance(
                                def_.atom.addr, atoms.SpOffset
                            ):
                                if to_replace.offset == def_.atom.addr.offset:
                                    defs.append(def_)
                        elif isinstance(to_replace, Register):
                            if isinstance(def_.atom, atoms.Register) and to_replace.reg_offset == def_.atom.reg_offset:
                                defs.append(def_)
                if len(defs) != 1:
                    continue
                the_def = defs[0]
            else:
                # find uses
                defs = rd.all_uses.get_uses_by_location(eq.codeloc)
                if len(defs) != 1:
                    # there are multiple defs for this register - we do not support replacing all of them
                    continue
                for def_ in defs:
                    def_: Definition
                    if isinstance(def_.atom, atoms.Register) and def_.atom.reg_offset == to_replace.reg_offset:
                        # found it!
                        the_def = def_
                        break
            if the_def is None:
                continue
            if the_def.codeloc.context:
                # the definition is in a callee function
                continue

            if isinstance(the_def.codeloc, ExternalCodeLocation):
                # this is a function argument. we enter a slightly different logic and try to eliminate copies of this
                # argument if
                # (a) the on-stack or in-register copy of it has never been modified in this function
                # (b) the function argument register has never been updated.
                #     TODO: we may loosen requirement (b) once we have real register versioning in AIL.
                defs = [def_ for def_ in rd.all_definitions if def_.codeloc == eq.codeloc]
                all_uses_with_def = None
                replace_with = None
                remove_initial_assignment = None

                if defs and len(defs) == 1:
                    arg_copy_def = defs[0]
                    if (
                        isinstance(arg_copy_def.atom, atoms.MemoryLocation)
                        and isinstance(arg_copy_def.atom.addr, SpOffset)
                        or isinstance(arg_copy_def.atom, atoms.Register)
                    ):
                        # found the copied definition (either a stack variable or a register variable)

                        # Make sure there is no other write to this stack location if the copy is a stack variable
                        if isinstance(arg_copy_def.atom, atoms.MemoryLocation):
                            if any(
                                (def_ != arg_copy_def and def_.atom == arg_copy_def.atom)
                                for def_ in rd.all_definitions
                                if isinstance(def_.atom, atoms.MemoryLocation)
                            ):
                                continue

                            # Make sure the register is never updated across this function
                            if any(
                                (def_ != the_def and def_.atom == the_def.atom)
                                for def_ in rd.all_definitions
                                if isinstance(def_.atom, atoms.Register) and rd.all_uses.get_uses(def_)
                            ):
                                continue

                        # find all its uses
                        all_arg_copy_var_uses: Set[Tuple[CodeLocation, Any]] = set(
                            rd.all_uses.get_uses_with_expr(arg_copy_def)
                        )
                        all_uses_with_def = set()

                        should_abort = False
                        for use in all_arg_copy_var_uses:
                            used_expr = use[1]
                            if used_expr is not None and used_expr.size != arg_copy_def.size:
                                should_abort = True
                                break
                            all_uses_with_def.add((arg_copy_def, use))
                        if should_abort:
                            continue

                        # to_replace = Load(None, StackBaseOffset(None, self.project.arch.bits, eq.atom0.offset),
                        #                  eq.atom0.size, endness=self.project.arch.memory_endness)
                        replace_with = eq.atom1
                        remove_initial_assignment = True

                if all_uses_with_def is None:
                    continue

            else:
                if (
                    eq.codeloc.block_addr == the_def.codeloc.block_addr
                    and eq.codeloc.block_idx == the_def.codeloc.block_idx
                ):
                    # the definition and the eq location are within the same block, and the definition is before
                    # the eq location.
                    if eq.codeloc.stmt_idx < the_def.codeloc.stmt_idx:
                        continue
                else:
                    # the definition is in the predecessor block of the eq
                    eq_block = next(
                        iter(
                            bb
                            for bb in self.func_graph
                            if bb.addr == eq.codeloc.block_addr and bb.idx == eq.codeloc.block_idx
                        )
                    )
                    eq_block_preds = set(self.func_graph.predecessors(eq_block))
                    if not any(
                        pred.addr == the_def.codeloc.block_addr and pred.idx == the_def.codeloc.block_idx
                        for pred in eq_block_preds
                    ):
                        continue

                if isinstance(eq.atom0, SimStackVariable):
                    # create the memory loading expression
                    new_idx = None if self._ail_manager is None else next(self._ail_manager.atom_ctr)
                    replace_with = Load(
                        new_idx,
                        StackBaseOffset(None, self.project.arch.bits, eq.atom0.offset),
                        eq.atom0.size,
                        endness=self.project.arch.memory_endness,
                    )
                elif isinstance(eq.atom0, SimMemoryVariable) and isinstance(eq.atom0.addr, int):
                    # create the memory loading expression
                    new_idx = None if self._ail_manager is None else next(self._ail_manager.atom_ctr)
                    replace_with = Load(
                        new_idx,
                        Const(None, None, eq.atom0.addr, self.project.arch.bits),
                        eq.atom0.size,
                        endness=self.project.arch.memory_endness,
                    )
                elif isinstance(eq.atom0, Register):
                    if isinstance(eq.atom1, Register):
                        if self.project.arch.is_artificial_register(eq.atom0.reg_offset, eq.atom0.size):
                            replace_with = eq.atom1
                        else:
                            replace_with = eq.atom0
                    else:
                        raise RuntimeError("Unsupported atom1 type %s." % type(eq.atom1))
                else:
                    raise RuntimeError("Unsupported atom0 type %s." % type(eq.atom0))

                to_replace_def = the_def

                # find all uses of this definition
                # we make a copy of the set since we may touch the set (uses) when replacing expressions
                all_uses: Set[Tuple[CodeLocation, Any]] = set(rd.all_uses.get_uses_with_expr(to_replace_def))
                # make sure none of these uses are phi nodes (depends on more than one def)
                all_uses_with_unique_def = set()
                for use_and_expr in all_uses:
                    use_loc, used_expr = use_and_expr
                    defs_and_exprs = rd.all_uses.get_uses_by_location(use_loc, exprs=True)
                    filtered_defs = {def_ for def_, expr_ in defs_and_exprs if expr_ == used_expr}
                    if len(filtered_defs) == 1:
                        all_uses_with_unique_def.add(use_and_expr)
                    else:
                        # optimization: break early
                        break

                if len(all_uses) != len(all_uses_with_unique_def):
                    # only when all uses are determined by the same definition will we continue with the simplification
                    continue

                # one more check: there can be at most one assignment in all these use locations
                assignment_ctr = 0
                for use_loc, used_expr in all_uses:
                    block = addr_and_idx_to_block[(use_loc.block_addr, use_loc.block_idx)]
                    stmt = block.statements[use_loc.stmt_idx]
                    if isinstance(stmt, Assignment):
                        assignment_ctr += 1
                if assignment_ctr > 1:
                    continue

                all_uses_with_def = {(to_replace_def, use_and_expr) for use_and_expr in all_uses}

                remove_initial_assignment = False  # expression folding will take care of it

            # ensure the uses we consider are all after the eq location
            filtered_all_uses_with_def = []
            for def_, use_and_expr in all_uses_with_def:
                u = use_and_expr[0]
                if (
                    u.block_addr == eq.codeloc.block_addr
                    and u.block_idx == eq.codeloc.block_idx
                    and u.stmt_idx < eq.codeloc.stmt_idx
                ):
                    # this use happens before the assignment - ignore it
                    continue
                filtered_all_uses_with_def.append((def_, use_and_expr))
            all_uses_with_def = filtered_all_uses_with_def

            if not all_uses_with_def:
                # definitions without uses may simply be our data-flow analysis being incorrect. do not remove them.
                continue

            # TODO: We can only replace all these uses with the stack variable if the stack variable isn't
            # TODO: re-assigned of a new value. Perform this check.

            # replace all uses
            all_uses_replaced = True
            for def_, use_and_expr in all_uses_with_def:
                u, used_expr = use_and_expr
                if u == eq.codeloc:
                    # skip the very initial assignment location
                    continue
                old_block = addr_and_idx_to_block.get((u.block_addr, u.block_idx), None)
                if old_block is None:
                    continue
                if used_expr is None:
                    all_uses_replaced = False
                    continue

                # ensure the expression that we want to replace with is still up-to-date
                replace_with_original_def = self._find_atom_def_at(replace_with, rd, def_.codeloc)
                if replace_with_original_def is not None and not self._check_atom_last_def(
                    replace_with, u, rd, replace_with_original_def
                ):
                    all_uses_replaced = False
                    continue

                # if there is an updated block, use it
                the_block = self.blocks.get(old_block, old_block)
                stmt: Statement = the_block.statements[u.stmt_idx]

                replace_with_copy = replace_with.copy()
                if used_expr.size != replace_with_copy.size:
                    new_idx = None if self._ail_manager is None else next(self._ail_manager.atom_ctr)
                    replace_with_copy = Convert(
                        new_idx,
                        replace_with_copy.bits,
                        used_expr.bits,
                        False,
                        replace_with_copy,
                    )

                r, new_block = self._replace_expr_and_update_block(
                    the_block, u.stmt_idx, stmt, def_, u, used_expr, replace_with_copy
                )
                if r:
                    self.blocks[old_block] = new_block
                else:
                    # failed to replace a use - we need to keep the initial assignment!
                    all_uses_replaced = False
                simplified |= r

            if all_uses_replaced and remove_initial_assignment:
                # the initial statement can be removed
                self._assignments_to_remove.add(eq.codeloc)

        if simplified:
            self._clear_cache()
        return simplified

    @staticmethod
    def _find_atom_def_at(atom, rd, codeloc: CodeLocation) -> Optional[Definition]:
        if isinstance(atom, Register):
            defs = rd.get_defs(atom, codeloc, OP_BEFORE)
            return next(iter(defs)) if len(defs) == 1 else None

        return None

    @staticmethod
    def _check_atom_last_def(atom, codeloc, rd, the_def) -> bool:
        if isinstance(atom, Register):
            defs = rd.get_defs(atom, codeloc, OP_BEFORE)
            for d in defs:
                if d.codeloc != the_def.codeloc:
                    return False

        return True

    #
    # Folding call expressions
    #

    @staticmethod
    def _is_expr_using_temporaries(expr: Expression) -> bool:
        walker = AILBlockTempCollector()
        walker.walk_expression(expr)
        return len(walker.temps) > 0

    @staticmethod
    def _is_stmt_using_temporaries(stmt: Statement) -> bool:
        walker = AILBlockTempCollector()
        walker.walk_statement(stmt)
        return len(walker.temps) > 0

    def _fold_call_exprs(self) -> bool:
        """
        Fold a call expression (statement) into other statements if the return value of the call expression (statement)
        is only used once, and the use site and the call site belongs to the same supernode.

        Example::

            s1 = func();
            s0 = s1;
            if (s0) ...

        after folding, it will be transformed to::

            s0 = func();
            if (s0) ...

        s0 can be folded into the condition, which means this example can further be transformed to::

            if (func()) ...

        this behavior is controlled by fold_callexprs_into_conditions. This to avoid cases where func() is called more
        than once after simplification and graph structuring where conditions might be duplicated (e.g., in Dream).
        In such cases, the one-use expression folder in RegionSimplifier will perform this transformation.
        """

        simplified = False

        prop = self._compute_propagation()
        if not prop.model.equivalence:
            return simplified

        addr_and_idx_to_block: Dict[Tuple[int, int], Block] = {}
        for block in self.func_graph.nodes():
            addr_and_idx_to_block[(block.addr, block.idx)] = block

        def_locations_to_remove: Set[CodeLocation] = set()
        updated_use_locations: Set[CodeLocation] = set()

        eq: Equivalence
        for eq in prop.model.equivalence:
            # register variable == Call
            if isinstance(eq.atom0, Register):
                call_addr: Optional[int]
                if isinstance(eq.atom1, Call):
                    # register variable = Call
                    call: Expression = eq.atom1
                    call_addr = call.target.value if isinstance(call.target, Const) else None
                elif isinstance(eq.atom1, Convert) and isinstance(eq.atom1.operand, Call):
                    # register variable = Convert(Call)
                    call = eq.atom1
                    call_addr = call.operand.target.value if isinstance(call.operand.target, Const) else None
                else:
                    continue

                if self._is_expr_using_temporaries(call):
                    continue

                if eq.codeloc in updated_use_locations:
                    # this def is now created by an updated use. the corresponding statement will be updated in the end.
                    # we must rerun Propagator to get an updated definition (and Equivalence)
                    continue

                # find the definition of this register
                rd = self._compute_reaching_definitions()
                defs = [
                    d
                    for d in rd.all_definitions
                    if d.codeloc == eq.codeloc
                    and isinstance(d.atom, atoms.Register)
                    and d.atom.reg_offset == eq.atom0.reg_offset
                ]
                if not defs or len(defs) > 1:
                    continue
                the_def: Definition = defs[0]

                # find all uses of this definition
                all_uses: Set[Tuple[CodeLocation, Any]] = set(rd.all_uses.get_uses_with_expr(the_def))

                if len(all_uses) != 1:
                    continue
                u, used_expr = next(iter(all_uses))
                if used_expr is None:
                    continue

                if u in def_locations_to_remove:
                    # this use site has been altered by previous folding attempts. the corresponding statement will be
                    # removed in the end. in this case, this Equivalence is probably useless, and we must rerun
                    # Propagator to get an updated Equivalence.
                    continue

                if not self._fold_callexprs_into_conditions:
                    # check the statement and make sure it's not a conditional jump
                    the_block = addr_and_idx_to_block[(u.block_addr, u.block_idx)]
                    if isinstance(the_block.statements[u.stmt_idx], ConditionalJump):
                        continue

                # check if the use and the definition is within the same supernode
                super_node_blocks = self._get_super_node_blocks(
                    addr_and_idx_to_block[(the_def.codeloc.block_addr, the_def.codeloc.block_idx)]
                )
                if u.block_addr not in {b.addr for b in super_node_blocks}:
                    continue

                # check if any atoms that the call relies on has been overwritten by statements in between the def site
                # and the use site.
                defsite_all_expr_uses = set(rd.all_uses.get_uses_by_location(the_def.codeloc))
                defsite_used_atoms = set()
                for dd in defsite_all_expr_uses:
                    defsite_used_atoms.add(dd.atom)
                usesite_expr_def_outdated = False
                for defsite_expr_atom in defsite_used_atoms:
                    usesite_expr_uses = set(rd.get_defs(defsite_expr_atom, u, OP_BEFORE))
                    if not usesite_expr_uses:
                        # the atom is not defined at the use site - it's fine
                        continue
                    defsite_expr_uses = set(rd.get_defs(defsite_expr_atom, the_def.codeloc, OP_BEFORE))
                    if usesite_expr_uses != defsite_expr_uses:
                        # special case: ok if this atom is assigned to at the def site and has not been overwritten
                        if len(usesite_expr_uses) == 1:
                            usesite_expr_use = next(iter(usesite_expr_uses))
                            if usesite_expr_use.atom == defsite_expr_atom and (
                                usesite_expr_use.codeloc == the_def.codeloc
                                or usesite_expr_use.codeloc.block_addr == call_addr
                            ):
                                continue
                        usesite_expr_def_outdated = True
                        break
                if usesite_expr_def_outdated:
                    continue

                # check if there are any calls in between the def site and the use site
                if self._count_calls_in_supernodeblocks(super_node_blocks, the_def.codeloc, u) > 0:
                    continue

                # replace all uses
                old_block = addr_and_idx_to_block.get((u.block_addr, u.block_idx), None)
                if old_block is None:
                    continue

                # if there is an updated block, use that
                the_block = self.blocks.get(old_block, old_block)
                stmt: Statement = the_block.statements[u.stmt_idx]

                if isinstance(eq.atom0, Register):
                    src = used_expr
                    dst: Expression = call

                    if src.bits != dst.bits:
                        dst = Convert(None, dst.bits, src.bits, False, dst)
                else:
                    continue

                # ensure what we are going to replace only appears once
                expr_ctr = SingleExpressionCounter(stmt, src)
                if expr_ctr.count > 1:
                    continue

                replaced, new_block = self._replace_expr_and_update_block(
                    the_block, u.stmt_idx, stmt, the_def, u, src, dst
                )

                if replaced:
                    self.blocks[old_block] = new_block
                    # this call has been folded to the use site. we can remove this call.
                    self._calls_to_remove.add(eq.codeloc)
                    simplified = True
                    def_locations_to_remove.add(eq.codeloc)
                    updated_use_locations.add(u)

        # no need to clear the cache at the end of this method
        return simplified

    def _get_super_node_blocks(self, start_node: Block) -> List[Block]:
        lst: List[Block] = [start_node]
        while True:
            b = lst[-1]
            successors = list(self.func_graph.successors(b))
            if len(successors) == 0:
                break
            if len(successors) == 1:
                succ = successors[0]
                # check its predecessors
                succ_predecessors = list(self.func_graph.predecessors(succ))
                if len(succ_predecessors) == 1:
                    lst.append(succ)
                else:
                    break
            else:
                # too many successors
                break
        return lst

    def _replace_expr_and_update_block(
        self, block, stmt_idx, stmt, the_def, codeloc, src_expr, dst_expr
    ) -> Tuple[bool, Optional[Block]]:
        replaced, new_stmt = stmt.replace(src_expr, dst_expr)
        if replaced:
            new_block = block.copy()
            new_block.statements = block.statements[::]
            new_block.statements[stmt_idx] = new_stmt

            # update the uses
            rd = self._compute_reaching_definitions()
            rd.all_uses.remove_use(the_def, codeloc, expr=src_expr)
            return True, new_block

        return False, None

    def _remove_dead_assignments(self) -> bool:
        stmts_to_remove_per_block: Dict[Tuple[int, int], Set[int]] = defaultdict(set)

        # Find all statements that should be removed
        mask = (1 << self.project.arch.bits) - 1

        rd = self._compute_reaching_definitions()
        stackarg_offsets = (
            {(tpl[1] & mask) for tpl in self._stack_arg_offsets} if self._stack_arg_offsets is not None else None
        )
        def_: Definition
        for def_ in rd.all_definitions:
            if def_.dummy:
                continue
            # we do not remove references to global memory regions no matter what
            if isinstance(def_.atom, atoms.MemoryLocation) and isinstance(def_.atom.addr, int):
                continue
            if isinstance(def_.atom, atoms.MemoryLocation):
                if not self._remove_dead_memdefs:
                    # we always remove definitions for stack arguments
                    if stackarg_offsets is not None and isinstance(def_.atom.addr, atoms.SpOffset):
                        if (def_.atom.addr.offset & mask) not in stackarg_offsets:
                            continue
                    else:
                        continue

            uses = rd.all_uses.get_uses(def_)

            if not uses:
                if not isinstance(def_.codeloc, ExternalCodeLocation):
                    stmts_to_remove_per_block[(def_.codeloc.block_addr, def_.codeloc.block_idx)].add(
                        def_.codeloc.stmt_idx
                    )

        for codeloc in self._calls_to_remove | self._assignments_to_remove:
            # this call can be removed. make sure it exists in stmts_to_remove_per_block
            stmts_to_remove_per_block[codeloc.block_addr, codeloc.block_idx].add(codeloc.stmt_idx)

        simplified = False

        # Remove the statements
        for old_block in self.func_graph.nodes():
            # if there is an updated block, use it
            block = self.blocks.get(old_block, old_block)

            if not isinstance(block, Block):
                continue

            if (block.addr, block.idx) not in stmts_to_remove_per_block:
                continue

            new_statements = []
            stmts_to_remove = stmts_to_remove_per_block[(block.addr, block.idx)]

            if not stmts_to_remove:
                continue

            for idx, stmt in enumerate(block.statements):
                if idx in stmts_to_remove:
                    if isinstance(stmt, (Assignment, Store)):
                        # Skip Assignment and Store statements
                        # if this statement triggers a call, it should only be removed if it's in self._calls_to_remove
                        codeloc = CodeLocation(block.addr, idx, ins_addr=stmt.ins_addr, block_idx=block.idx)
                        if codeloc in self._assignments_to_remove:
                            # it should be removed
                            simplified = True
                            continue

                        if self._statement_has_call_exprs(stmt):
                            if codeloc in self._calls_to_remove:
                                # it has a call and must be removed
                                simplified = True
                                continue
                        else:
                            # no calls. remove it
                            simplified = True
                            continue
                    elif isinstance(stmt, Call):
                        codeloc = CodeLocation(block.addr, idx, ins_addr=stmt.ins_addr, block_idx=block.idx)
                        if codeloc in self._calls_to_remove:
                            # this call can be removed
                            simplified = True
                            continue

                        if stmt.ret_expr is not None:
                            # the return expr is not used. it should not have return expr
                            if stmt.fp_ret_expr is not None:
                                # maybe its fp_ret_expr is used?
                                stmt = stmt.copy()
                                stmt.ret_expr = stmt.fp_ret_expr
                                stmt.fp_ret_expr = None
                            else:
                                # clear ret_expr
                                stmt = stmt.copy()
                                stmt.ret_expr = None
                                simplified = True
                    else:
                        # Should not happen!
                        raise NotImplementedError()

                new_statements.append(stmt)

            new_block = block.copy()
            new_block.statements = new_statements
            self.blocks[old_block] = new_block

        return simplified

    #
    # Rewriting ccalls
    #

    def _rewrite_ccalls(self):
        rewriter_cls = CCALL_REWRITERS.get(self.project.arch.name, None)
        if rewriter_cls is None:
            return False

        walker = None

        class _any_update:
            """
            Dummy class for storing if any result has been updated.
            """

            v = False

        def _handle_expr(
            expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement, block
        ) -> Optional[Expression]:
            if isinstance(expr, DirtyExpression) and isinstance(expr.dirty_expr, VEXCCallExpression):
                rewriter = rewriter_cls(expr.dirty_expr, self.project.arch)
                if rewriter.result is not None:
                    _any_update.v = True
                    return rewriter.result
                else:
                    return None

            return AILBlockWalker._handle_expr(walker, expr_idx, expr, stmt_idx, stmt, block)

        blocks_by_addr_and_idx = {(node.addr, node.idx): node for node in self.func_graph.nodes()}

        walker = AILBlockWalker()
        walker._handle_expr = _handle_expr

        updated = False
        for block in blocks_by_addr_and_idx.values():
            _any_update.v = False
            old_block = block.copy()
            walker.walk(block)
            if _any_update.v:
                self.blocks[old_block] = block
                updated = True

        return updated

    @staticmethod
    def _statement_has_call_exprs(stmt: Statement) -> bool:
        def _handle_callexpr(expr_idx, expr, stmt_idx, stmt, block):  # pylint:disable=unused-argument
            raise HasCallNotification()

        walker = AILBlockWalker()
        walker.expr_handlers[Call] = _handle_callexpr
        try:
            walker.walk_statement(stmt)
        except HasCallNotification:
            return True

        return False

    @staticmethod
    def _expression_has_call_exprs(expr: Expression) -> bool:
        def _handle_callexpr(expr_idx, expr, stmt_idx, stmt, block):  # pylint:disable=unused-argument
            raise HasCallNotification()

        walker = AILBlockWalker()
        walker.expr_handlers[Call] = _handle_callexpr
        try:
            walker.walk_expression(expr)
        except HasCallNotification:
            return True

        return False

    @staticmethod
    def _count_calls_in_supernodeblocks(blocks: List[Block], start: CodeLocation, end: CodeLocation) -> int:
        """
        Count the number of call statements in a list of blocks for a single super block between two given code
        locations (exclusive).
        """
        calls = 0
        started = False
        for b in blocks:
            if b.addr == start.block_addr:
                started = True
                continue
            if b.addr == end.block_addr:
                started = False
                continue

            if started:
                if b.statements and isinstance(b.statements[-1], Call):
                    calls += 1
        return calls


AnalysesHub.register_default("AILSimplifier", AILSimplifier)



import angr


class MessageBoxA(angr.SimProcedure):
    def run(self, hWnd, lpText, lpCaption, uType):
        text = self.extract(lpText)
        if not self.state.solver.is_true(lpCaption == 0):
            caption = self.extract(lpCaption)
        else:
            caption = "Error"

        result = self.state.solver.If(
            uType & 0xF == 0, 1, self.state.solver.BVS("messagebox_button", 32, key=("api", "messagebox", "button"))
        )
        self.state.history.add_event("message_box", text=text, caption=caption, result=result)
        return result

    def extract(self, addr):
        return self.state.mem[addr].string.concrete


class MessageBoxExA(MessageBoxA):
    def run(self, hWnd, lpText, lpCaption, uType, wLanguageId):
        super().run(hWnd, lpText, lpCaption, uType)


class MessageBoxW(MessageBoxA):
    def extract(self, addr):
        return self.state.mem[addr].wstring.concrete


class MessageBoxExW(MessageBoxW, MessageBoxExA):
    pass


class MessageBoxIndirectA(MessageBoxExA):
    def run(self, lpMsgBoxParams):
        if self.arch.bits != 32:
            raise angr.errors.SimProcedureError("MessageBoxIndirectA is only implemented for 32 bit windows")
        hwndOwner = self.state.mem[lpMsgBoxParams + 0x4].dword.resolved
        lpszText = self.state.mem[lpMsgBoxParams + 0xC].dword.resolved
        lpszCaption = self.state.mem[lpMsgBoxParams + 0x10].dword.resolved
        dwStyle = self.state.mem[lpMsgBoxParams + 0x14].dword.resolved
        dwLanguageId = self.state.mem[lpMsgBoxParams + 0x24].dword.resolved
        super().run(hwndOwner, lpszText, lpszCaption, dwStyle, dwLanguageId)


class MessageBoxIndirectW(MessageBoxW, MessageBoxIndirectA):
    pass



#!/usr/bin/env python3
import networkx
import unittest

from angr.analyses.cfg_slice_to_sink import CFGSliceToSink, slice_callgraph, slice_cfg_graph, slice_function_graph


class _MockCFGNode:
    def __init__(self, addr):
        self.addr = addr

    def __repr__(self):
        return "%s" % self.addr


def _a_graph_and_its_nodes():
    # Build the following graph (addresses displayed):
    # 0 -> 1, 1 -> 2, 0 -> 3
    graph = networkx.DiGraph()
    nodes = list(map(_MockCFGNode, range(4)))
    graph.add_edge(nodes[0], nodes[1])
    graph.add_edge(nodes[1], nodes[2])
    graph.add_edge(nodes[0], nodes[3])
    return (graph, nodes)


class TestGraph(unittest.TestCase):
    def test_slice_callgraph_remove_content_not_in_a_cfg_slice_to_sink(self):
        my_callgraph, nodes = _a_graph_and_its_nodes()

        # Let's imagine a node (0x42), not a function entry point, not in my_callgraph, such as:
        # 1 -> 0x42, 0x42 -> 2
        transitions = {nodes[0]: [nodes[1]], nodes[1]: [0x42], 0x42: [nodes[2]]}
        cfg_slice_to_sink = CFGSliceToSink(None, transitions)

        sliced_callgraph = slice_callgraph(my_callgraph, cfg_slice_to_sink)

        result_nodes = list(sliced_callgraph.nodes)
        result_edges = list(sliced_callgraph.edges)

        self.assertListEqual(result_nodes, [nodes[0], nodes[1], nodes[2]])
        self.assertListEqual(result_edges, [(nodes[0], nodes[1]), (nodes[1], nodes[2])])

    def test_slice_callgraph_mutates_the_original_graph(self):
        my_callgraph, nodes = _a_graph_and_its_nodes()

        # Let's imagine a node (0x42), not a function entry point, not in my_callgraph, such as:
        # 1 -> 0x42, 0x42 -> 2
        transitions = {nodes[0]: [nodes[1]], nodes[1]: [0x42], 0x42: [nodes[2]]}
        cfg_slice_to_sink = CFGSliceToSink(None, transitions)

        sliced_callgraph = slice_callgraph(my_callgraph, cfg_slice_to_sink)

        self.assertEqual(len(my_callgraph.nodes), 3)
        self.assertEqual(len(my_callgraph.edges), 2)
        self.assertEqual(my_callgraph, sliced_callgraph)

    def test_slice_cfg_graph_remove_content_not_in_a_cfg_slice_to_sink(self):
        my_graph, nodes = _a_graph_and_its_nodes()

        transitions = {nodes[0].addr: [nodes[1].addr], nodes[1].addr: [nodes[2].addr]}
        my_slice = CFGSliceToSink(None, transitions)

        sliced_graph = slice_cfg_graph(my_graph, my_slice)
        result_nodes = list(sliced_graph.nodes)
        result_edges = list(sliced_graph.edges)

        self.assertListEqual(result_nodes, [nodes[0], nodes[1], nodes[2]])
        self.assertListEqual(result_edges, [(nodes[0], nodes[1]), (nodes[1], nodes[2])])

    def test_slice_cfg_graph_mutates_the_original_graph(self):
        my_graph, nodes = _a_graph_and_its_nodes()

        transitions = {nodes[0].addr: [nodes[1].addr]}
        my_slice = CFGSliceToSink(None, transitions)

        sliced_graph = slice_cfg_graph(my_graph, my_slice)

        self.assertEqual(len(my_graph.nodes), 2)
        self.assertEqual(len(my_graph.edges), 1)
        self.assertEqual(my_graph, sliced_graph)

    def test_slice_function_graph_remove_nodes_not_in_a_cfg_slice_to_sink(self):
        # Imagine a CFG being:    0 -> 0x42, 0x42 -> 1, 1 -> 2, 0 -> 3
        # And the function graph: 0 -> 1, 1 -> 2, 0 -> 3
        my_function_graph, nodes = _a_graph_and_its_nodes()

        transitions = {nodes[0].addr: [0x42], 0x42: [nodes[1].addr]}
        my_slice = CFGSliceToSink(None, transitions)

        sliced_function_graph = slice_function_graph(my_function_graph, my_slice)
        result_nodes = list(sliced_function_graph.nodes)
        result_edges = list(sliced_function_graph.edges)

        self.assertListEqual(result_nodes, [nodes[0], nodes[1]])
        self.assertListEqual(result_edges, [(nodes[0], nodes[1])])

    def test_slice_function_graph_mutates_the_original_function_graph(self):
        # Imagine a CFG being:    0 -> 0x42, 0x42 -> 1, 1 -> 2, 0 -> 3
        # And the function graph: 0 -> 1, 1 -> 2, 0 -> 3
        my_function_graph, nodes = _a_graph_and_its_nodes()

        transitions = {nodes[0].addr: [0x42], 0x42: [nodes[1].addr]}
        my_slice = CFGSliceToSink(None, transitions)

        sliced_function_graph = slice_function_graph(my_function_graph, my_slice)

        self.assertEqual(len(my_function_graph.nodes), 2)
        self.assertEqual(len(my_function_graph.edges), 1)
        self.assertEqual(my_function_graph, sliced_function_graph)


if __name__ == "__main__":
    unittest.main()



import logging

from .base import SimSootStmt

l = logging.getLogger("angr.engines.soot.statements.goto")


class SimSootStmt_Goto(SimSootStmt):
    def _execute(self):
        jmp_target = self._get_bb_addr_from_instr(instr=self.stmt.target)
        self._add_jmp_target(target=jmp_target, condition=self.state.solver.true)



#!/usr/bin/env python3
__package__ = __package__ or "tests.analyses.cfg"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestCfgThumbFirmware(unittest.TestCase):
    def test_thumb_firmware_cfg(self):
        # Test an ARM firmware sample.
        #
        # This tests CFG, but also the Gym (the ThumbSpotter, etc)
        # Also requires proper relocs support, or You're Gonna Have a Bad Time(tm)
        # In short, a very comprehensive high level test

        path = os.path.join(test_location, "armel", "i2c_master_read-nucleol152re.elf")
        p = angr.Project(path, auto_load_libs=False)

        # This is the canonical way to carve up a nasty firmware thing.

        cfg = p.analyses.CFGFast(
            resolve_indirect_jumps=True, force_smart_scan=False, force_complete_scan=False, normalize=True
        )

        # vfprintf should return; this function has a weird C++ thing that gets compiled as a tail-call.
        # The function itself must return, and _NOT_ contain its callee.
        vfprintf = cfg.kb.functions[p.loader.find_symbol("vfprintf").rebased_addr]
        assert vfprintf.returning
        assert len(list(vfprintf.blocks)) == 1
        # The function should have one "transition"
        block = list(vfprintf.endpoints_with_type["transition"])[0]
        assert len(block.successors()) == 1
        succ = list(block.successors())[0]
        assert succ.addr == 0x080081DD
        f2 = p.kb.functions[succ.addr]
        assert f2.name == "_vfprintf_r"
        assert f2.returning


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,disable=no-self-use
__package__ = __package__ or "tests.state_plugins.posix"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestFileStructFuncs(unittest.TestCase):
    def check_state_1(self, state):
        # Need to dump file.txt by path because program closes it
        return (
            state.posix.dump_file_by_path("file.txt") == b"testing abcdef"
            and state.posix.dumps(0)[:4] == b"xyz\n"
            and state.posix.dumps(1) == b"good1\n"
            and state.posix.dumps(2) == b""
        )

    def check_state_2(self, state):
        return (
            state.posix.dump_file_by_path("file.txt") == b"testing abcdef"
            and state.posix.dumps(0)[:4] == b"wxyz"
            and state.posix.dumps(1) == b""
            and state.posix.dumps(2) == b"good2\n"
        )

    def check_state_3(self, state):
        return (
            state.posix.dump_file_by_path("file.txt") == b"testing abcdef"
            and state.posix.dumps(1) == b""
            and state.posix.dumps(2) == b""
        )

    def test_file_struct_funcs(self):
        test_bin = os.path.join(test_location, "x86_64", "file_func_test")
        b = angr.Project(test_bin, auto_load_libs=False)

        pg = b.factory.simulation_manager()
        pg.active[0].options.discard("LAZY_SOLVES")
        pg.explore()

        assert len(pg.deadended) == 3

        for p in pg.deadended:
            assert self.check_state_1(p) or self.check_state_2(p) or self.check_state_3(p)


if __name__ == "__main__":
    unittest.main()



import logging
from typing import Dict, Tuple

from ..calling_conventions import SYSCALL_CC, SimCCSyscall
from ..errors import AngrUnsupportedSyscallError, SimSolverError
from ..procedures import SIM_PROCEDURES as P
from .simos import SimOS

_l = logging.getLogger(name=__name__)


class SimUserland(SimOS):
    """
    This is a base class for any SimOS that wants to support syscalls.

    It uses the CLE kernel object to provide addresses for syscalls. Syscalls will be emulated as a jump to one of these
    addresses, where a SimProcedure from the syscall library provided at construction time will be executed.
    """

    def __init__(self, project, syscall_library=None, syscall_addr_alignment=4, **kwargs):
        super().__init__(project, **kwargs)
        self.syscall_library = syscall_library.copy()
        self.syscall_addr_alignment = syscall_addr_alignment
        self.kernel_base = None
        self.unknown_syscall_number = None
        self.syscall_abis: Dict[str, Tuple[int, int, int]] = {}
        # syscall_abis is a dict of tuples {name: (base_number, min_number, max_number)}
        # min_number and max_number are just cached from SimSyscallLibrary.{min,max}imum_sysall_number
        # base_number is used to map the syscalls into the syscall address space - it's a "base address"
        # but a number. to convert from syscall number to address it's
        # (number - min_num + base_num) * alignment + kernel_base

    def configure_project(self, abi_list=None):  # pylint: disable=arguments-differ
        if abi_list is None:
            abi_list = list(self.syscall_library.syscall_number_mapping)
            assert (
                len(abi_list) == 1
            ), "More than one ABI is available for this target - you need to specify which ones are valid"
        self.kernel_base = self.project.loader.kernel_object.mapped_base

        base_no = 0
        for abi in abi_list:
            assert abi in self.syscall_library.syscall_number_mapping
            min_no = self.syscall_library.minimum_syscall_number(abi)
            max_no = self.syscall_library.maximum_syscall_number(abi)
            self.syscall_abis[abi] = (base_no, min_no, max_no)
            base_no += max_no - min_no + 1  # since max is the actual max and not the array length

        self.unknown_syscall_number = base_no

        # configure_project() involves lightweight symbolic execution, which may ends up using syscall ABIs. hence, we
        # need to fill in self.syscall_abis before calling configure_project().
        super().configure_project()

    def syscall_cc(self, state) -> SimCCSyscall:
        if state.os_name in SYSCALL_CC[state.arch.name]:
            cc = SYSCALL_CC[state.arch.name][state.os_name](state.arch)
        else:
            # Use the default syscall calling convention - it may bring problems
            _l.warning("No syscall calling convention available for %s/%s", state.arch.name, state.os_name)
            cc = SYSCALL_CC[state.arch.name]["default"](state.arch)
        return cc

    def syscall(self, state, allow_unsupported=True):
        """
        Given a state, return the procedure corresponding to the current syscall.
        This procedure will have .syscall_number, .display_name, and .addr set.

        :param state:               The state to get the syscall number from
        :param allow_unsupported:   Whether to return a "dummy" sycall instead of raising an unsupported exception
        """
        abi = self.syscall_abi(state)
        cc = self.syscall_cc(state)

        sym_num = cc.syscall_num(state)
        try:
            num = state.solver.eval_one(sym_num)
        except SimSolverError:
            if allow_unsupported:
                num = self.unknown_syscall_number
            else:
                if not state.solver.satisfiable():
                    raise AngrUnsupportedSyscallError("The program state is not satisfiable")
                else:
                    raise AngrUnsupportedSyscallError("Got a symbolic syscall number")

        proc = self.syscall_from_number(num, allow_unsupported=allow_unsupported, abi=abi)
        proc.cc = cc
        return proc

    def syscall_abi(self, state):  # pylint: disable=unused-argument,no-self-use
        """
        Optionally, override this function to determine which abi is being used for the state's current syscall.
        """
        return None

    def is_syscall_addr(self, addr):
        """
        Return whether or not the given address corresponds to a syscall implementation.
        """
        if self.kernel_base is None or addr < self.kernel_base:
            return False

        addr -= self.kernel_base

        if addr % self.syscall_addr_alignment != 0:
            return False

        addr //= self.syscall_addr_alignment
        return addr <= self.unknown_syscall_number

    def syscall_from_addr(self, addr, allow_unsupported=True):
        """
        Get a syscall SimProcedure from an address.

        :param addr: The address to convert to a syscall SimProcedure
        :param allow_unsupported: Whether to return a dummy procedure for an unsupported syscall instead of raising an
                                  exception.
        :return: The SimProcedure for the syscall, or None if the address is not a syscall address.
        """
        if not self.is_syscall_addr(addr):
            return None

        number = (addr - self.kernel_base) // self.syscall_addr_alignment
        for abi in self.syscall_abis:
            baseno, minno, maxno = self.syscall_abis[abi]
            if baseno <= number <= baseno + maxno - minno:
                number += minno
                number -= baseno
                break
        else:
            abi = None
        return self.syscall_from_number(number, allow_unsupported=allow_unsupported, abi=abi)

    def syscall_from_number(self, number, allow_unsupported=True, abi=None):
        """
        Get a syscall SimProcedure from its number.

        :param number:              The syscall number
        :param allow_unsupported:   Whether to return a "stub" syscall for unsupported numbers instead of throwing an
                                    error
        :param abi:                 The name of the abi to use. If None, will assume that the abis have disjoint
                                    numbering schemes and pick the right one.
        :return: The SimProcedure for the syscall
        """
        abilist = self.syscall_abis if abi is None else [abi]

        if self.syscall_library is None:
            if not allow_unsupported:
                raise AngrUnsupportedSyscallError("%s does not have a library of syscalls implemented" % self.name)
            proc = P["stubs"]["syscall"]()
        elif not allow_unsupported and not self.syscall_library.has_implementation(number, self.arch, abilist):
            raise AngrUnsupportedSyscallError("No implementation for syscall %d" % number)
        else:
            proc = self.syscall_library.get(number, self.arch, abilist)

        if proc.abi is not None:
            baseno, minno, _ = self.syscall_abis[proc.abi]
            mapno = number - minno + baseno
        else:
            mapno = self.unknown_syscall_number

        proc.addr = mapno * self.syscall_addr_alignment + self.kernel_base
        return proc



# pylint:disable=wrong-import-position,arguments-differ
import logging
from typing import List, Optional, Tuple

import pyvex
from pyvex import IRSB
from archinfo import ArchARM

try:
    from .engines import pcode
except ImportError:
    pcode = None

from .protos import primitives_pb2 as pb2
from .serializable import Serializable
from .engines.vex import VEXLifter

l = logging.getLogger(name=__name__)

DEFAULT_VEX_ENGINE = VEXLifter(None)  # this is only used when Block is not initialized with a project


class DisassemblerBlock:
    """
    Helper class to represent a block of dissassembled target architecture
    instructions
    """

    __slots__ = ["addr", "insns", "thumb", "arch"]

    def __init__(self, addr, insns, thumb, arch):
        self.addr = addr
        self.insns = insns
        self.thumb = thumb
        self.arch = arch

    def pp(self):
        print(str(self))

    def __str__(self):
        return "\n".join(map(str, self.insns))

    def __repr__(self):
        return "<DisassemblerBlock for %#x>" % self.addr


class DisassemblerInsn:
    """
    Helper class to represent a disassembled target architecture instruction
    """

    __slots__ = ()

    @property
    def size(self) -> int:
        raise NotImplementedError()

    @property
    def address(self) -> int:
        raise NotImplementedError()

    @property
    def mnemonic(self) -> str:
        raise NotImplementedError()

    @property
    def op_str(self) -> str:
        raise NotImplementedError()

    def __str__(self):
        return f"{self.address:#x}:\t{self.mnemonic}\t{self.op_str}"

    def __repr__(self):
        return f'<DisassemblerInsn "{self.mnemonic}" for {self.address:#x}>'


class CapstoneBlock(DisassemblerBlock):
    """
    Deep copy of the capstone blocks, which have serious issues with having extended lifespans
    outside of capstone itself
    """

    __slots__ = ()


class CapstoneInsn(DisassemblerInsn):
    """
    Represents a capstone instruction.
    """

    __slots__ = ("insn",)

    def __init__(self, capstone_insn):
        self.insn = capstone_insn

    @property
    def size(self) -> int:
        return self.insn.size

    @property
    def address(self) -> int:
        return self.insn.address

    @property
    def mnemonic(self) -> str:
        return self.insn.mnemonic

    @property
    def op_str(self) -> str:
        return self.insn.op_str

    def __getattr__(self, item):
        if item in ("__str__", "__repr__"):
            return self.__getattribute__(item)
        if hasattr(self.insn, item):
            return getattr(self.insn, item)
        raise AttributeError()


class Block(Serializable):
    """
    Represents a basic block in a binary or a program.
    """

    BLOCK_MAX_SIZE = 4096

    __slots__ = [
        "_project",
        "_bytes",
        "_vex",
        "thumb",
        "_disassembly",
        "_capstone",
        "addr",
        "size",
        "arch",
        "_instructions",
        "_instruction_addrs",
        "_opt_level",
        "_vex_nostmt",
        "_collect_data_refs",
        "_strict_block_end",
        "_cross_insn_opt",
        "_load_from_ro_regions",
        "_initial_regs",
    ]

    def __init__(
        self,
        addr,
        project=None,
        arch=None,
        size=None,
        byte_string=None,
        vex=None,
        thumb=False,
        backup_state=None,
        extra_stop_points=None,
        opt_level=None,
        num_inst=None,
        traceflags=0,
        strict_block_end=None,
        collect_data_refs=False,
        cross_insn_opt=True,
        load_from_ro_regions=False,
        initial_regs=None,
    ):
        # set up arch
        if project is not None:
            self.arch = project.arch
        else:
            self.arch = arch

        if self.arch is None:
            raise ValueError('Either "project" or "arch" has to be specified.')

        if project is not None and backup_state is None and project.kb.patches.values():
            backup_state = project.kb.patches.patched_entry_state

        if isinstance(self.arch, ArchARM):
            if addr & 1 == 1:
                thumb = True
            elif thumb:
                addr |= 1
        else:
            thumb = False

        self._project = project
        self.thumb = thumb
        self.addr = addr
        self._opt_level = opt_level
        self._initial_regs: Optional[List[Tuple[int, int, int]]] = initial_regs if collect_data_refs else None

        if self._project is None and byte_string is None:
            raise ValueError('"byte_string" has to be specified if "project" is not provided.')

        if size is None:
            if byte_string is not None:
                size = len(byte_string)
            elif vex is not None:
                size = vex.size
            else:
                if self._initial_regs:
                    self.set_initial_regs()
                vex = self._vex_engine.lift_vex(
                    clemory=project.loader.memory,
                    state=backup_state,
                    insn_bytes=byte_string,
                    addr=addr,
                    thumb=thumb,
                    extra_stop_points=extra_stop_points,
                    opt_level=opt_level,
                    num_inst=num_inst,
                    traceflags=traceflags,
                    strict_block_end=strict_block_end,
                    collect_data_refs=collect_data_refs,
                    load_from_ro_regions=load_from_ro_regions,
                    cross_insn_opt=cross_insn_opt,
                )
                if self._initial_regs:
                    self.reset_initial_regs()
                size = vex.size

        self._vex = vex
        self._vex_nostmt = None
        self._disassembly = None
        self._capstone = None
        self.size = size
        self._collect_data_refs = collect_data_refs
        self._strict_block_end = strict_block_end
        self._cross_insn_opt = cross_insn_opt
        self._load_from_ro_regions = load_from_ro_regions

        self._instructions = num_inst
        self._instruction_addrs: List[int] = []

        self._parse_vex_info(self._vex)

        if byte_string is None:
            if backup_state is not None:
                buffer, _, offset = self._vex_engine._load_bytes(addr - thumb, size, state=backup_state)
                self._bytes = buffer[offset:]
                if type(self._bytes) is memoryview:
                    self._bytes = bytes(self._bytes)
                elif type(self._bytes) is not bytes:
                    self._bytes = bytes(pyvex.ffi.buffer(self._bytes, size))
            else:
                self._bytes = None
        elif type(byte_string) is bytes:
            if self.size is not None:
                self._bytes = byte_string[: self.size]
            else:
                self._bytes = byte_string
        else:
            # Convert bytestring to a str
            # size will ALWAYS be known at this point
            self._bytes = str(pyvex.ffi.buffer(byte_string, self.size))

    def _parse_vex_info(self, vex_block):
        if vex_block is not None:
            self._instructions = vex_block.instructions
            self._instruction_addrs = vex_block.instruction_addresses
            self.size = vex_block.size

    def __repr__(self):
        return "<Block for %#x, %d bytes>" % (self.addr, self.size)

    def __getstate__(self):
        return {k: getattr(self, k) for k in self.__slots__ if k not in {"_capstone", "_disassembly", "_project"}}

    def __setstate__(self, data):
        for k, v in data.items():
            setattr(self, k, v)
        self._capstone = None
        self._disassembly = None
        self._project = None

    def __hash__(self):
        return hash((type(self), self.addr, self.bytes))

    def __eq__(self, other):
        return type(self) is type(other) and self.addr == other.addr and self.bytes == other.bytes

    def __ne__(self, other):
        return not self == other

    def pp(self, **kwargs):
        if self._project is not None:
            addr = self.addr - 1 if self.thumb else self.addr
            print(
                self._project.analyses.Disassembly(
                    ranges=[(addr, addr + self.size)],
                    thumb=self.thumb,
                    block_bytes=self.bytes,
                ).render(**kwargs)
            )
        else:
            self.disassembly.pp()

    def set_initial_regs(self):
        # for data reference collection, on some architectures, we need to set up initial registers
        if self._initial_regs is not None:
            for offset, size, value in self._initial_regs:  # pylint:disable=not-an-iterable
                pyvex.pvc.register_initial_register_value(offset, size, value)

    @staticmethod
    def reset_initial_regs():
        pyvex.pvc.reset_initial_register_values()

    @property
    def _vex_engine(self):
        if self._project is None:
            return DEFAULT_VEX_ENGINE
        else:
            return self._project.factory.default_engine

    @property
    def vex(self) -> IRSB:
        if not self._vex:
            if self._initial_regs:
                self.set_initial_regs()
            self._vex = self._vex_engine.lift_vex(
                clemory=self._project.loader.memory if self._project is not None else None,
                insn_bytes=self._bytes,
                addr=self.addr,
                thumb=self.thumb,
                size=self.size,
                num_inst=self._instructions,
                opt_level=self._opt_level,
                arch=self.arch,
                collect_data_refs=self._collect_data_refs,
                strict_block_end=self._strict_block_end,
                cross_insn_opt=self._cross_insn_opt,
                load_from_ro_regions=self._load_from_ro_regions,
            )
            if self._initial_regs:
                self.reset_initial_regs()
            self._parse_vex_info(self._vex)

        return self._vex

    @property
    def vex_nostmt(self):
        if self._vex_nostmt:
            return self._vex_nostmt

        if self._vex:
            return self._vex

        if self._initial_regs:
            self.set_initial_regs()
        self._vex_nostmt = self._vex_engine.lift_vex(
            clemory=self._project.loader.memory if self._project is not None else None,
            insn_bytes=self._bytes,
            addr=self.addr,
            thumb=self.thumb,
            size=self.size,
            num_inst=self._instructions,
            opt_level=self._opt_level,
            arch=self.arch,
            skip_stmts=True,
            collect_data_refs=self._collect_data_refs,
            strict_block_end=self._strict_block_end,
            cross_insn_opt=self._cross_insn_opt,
            load_from_ro_regions=self._load_from_ro_regions,
        )
        if self._initial_regs:
            self.reset_initial_regs()
        self._parse_vex_info(self._vex_nostmt)
        return self._vex_nostmt

    @property
    def _using_pcode_engine(self) -> bool:
        return (pcode is not None) and isinstance(self._vex_engine, pcode.HeavyPcodeMixin)

    @property
    def disassembly(self) -> DisassemblerBlock:
        """
        Provide a disassembly object using whatever disassembler is available
        """
        if self._disassembly is None:
            if self._using_pcode_engine:
                self._disassembly = self.vex.disassembly
            else:
                self._disassembly = self.capstone
        return self._disassembly

    @property
    def capstone(self):
        if self._capstone:
            return self._capstone

        cs = self.arch.capstone if not self.thumb else self.arch.capstone_thumb

        insns = []

        block_bytes = self.bytes
        if self.size is not None:
            block_bytes = block_bytes[: self.size]
        for cs_insn in cs.disasm(block_bytes, self.addr):
            insns.append(CapstoneInsn(cs_insn))
        block = CapstoneBlock(self.addr, insns, self.thumb, self.arch)

        self._capstone = block
        return block

    @property
    def codenode(self):
        return BlockNode(self.addr, self.size, bytestr=self.bytes, thumb=self.thumb)

    @property
    def bytes(self):
        if self._bytes is None:
            addr = self.addr
            if self.thumb:
                addr = (addr >> 1) << 1
            self._bytes = self._project.loader.memory.load(addr, self.size)
        return self._bytes

    @property
    def instructions(self):
        if not self._instructions and self._vex is None:
            # initialize from VEX
            _ = self.vex

        return self._instructions

    @property
    def instruction_addrs(self):
        if not self._instruction_addrs and self._vex is None:
            # initialize instruction addrs
            _ = self.vex

        return self._instruction_addrs

    @classmethod
    def _get_cmsg(cls):
        return pb2.Block()

    def serialize_to_cmessage(self):
        obj = self._get_cmsg()
        obj.ea = self.addr
        obj.size = self.size
        obj.bytes = self.bytes

        return obj

    @classmethod
    def parse_from_cmessage(cls, cmsg):
        obj = cls(
            cmsg.ea,
            size=cmsg.size,
            byte_string=cmsg.bytes,
        )
        return obj


class SootBlock:
    """
    Represents a Soot IR basic block.
    """

    def __init__(self, addr, project=None, arch=None):
        self.addr = addr
        self.arch = arch
        self._project = project
        self._the_binary = project.loader.main_object

    @property
    def _soot_engine(self):
        if self._project is None:
            raise Exception("SHIIIIIIIT")
        return self._project.factory.default_engine

    @property
    def soot(self):
        return self._soot_engine.lift_soot(self.addr, the_binary=self._the_binary)

    @property
    def size(self):
        stmts = None if self.soot is None else self.soot.statements
        stmts_len = len(stmts) if stmts else 0
        return stmts_len

    @property
    def codenode(self):
        stmts = None if self.soot is None else self.soot.statements
        stmts_len = len(stmts) if stmts else 0
        return SootBlockNode(self.addr, stmts_len, stmts=stmts)


from .codenode import BlockNode, SootBlockNode



#!/usr/bin/env python3
__package__ = __package__ or "tests.factory"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestArgc(unittest.TestCase):
    def test_mips(self):
        proj = angr.Project(os.path.join(test_location, "mips", "argc_decide"), auto_load_libs=False)
        r_addr = 0x4006F4

        s = proj.factory.entry_state(args=["aaa"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 1

        s = proj.factory.entry_state(args=["aaa", "bbb"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 0

    def test_mipsel(self):
        proj = angr.Project(os.path.join(test_location, "mipsel", "argc_decide"), auto_load_libs=False)
        r_addr = 0x400708
        s = proj.factory.entry_state(args=["aaa", "bbb"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 1

        s = proj.factory.entry_state(args=["aaa"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 0

    def test_i386(self):
        proj = angr.Project(os.path.join(test_location, "i386", "argc_decide"), auto_load_libs=False)
        r_addr = 0x80483D4
        s = proj.factory.entry_state(args=["aaa"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 1

        s = proj.factory.entry_state(args=["aaa", "bbb"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 0

    def test_amd64(self):
        proj = angr.Project(os.path.join(test_location, "x86_64", "argc_decide"), auto_load_libs=False)
        r_addr = 0x4004C7
        s = proj.factory.entry_state(args=["aaa"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 1

        s = proj.factory.entry_state(args=["aaa", "bbb"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 0

    def test_arm(self):
        proj = angr.Project(os.path.join(test_location, "armel", "argc_decide"), auto_load_libs=False)
        r_addr = 0x1040C

        s = proj.factory.entry_state(args=["aaa"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 1

        s = proj.factory.entry_state(args=["aaa", "bbb"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 0

    def test_ppc32(self):
        proj = angr.Project(os.path.join(test_location, "ppc", "argc_decide"), auto_load_libs=False)
        r_addr = 0x10000404

        s = proj.factory.entry_state(args=["aaa"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 1

        s = proj.factory.entry_state(args=["aaa", "bbb"], env={"HOME": "/home/angr"})
        xpl = proj.factory.simulation_manager(s).explore(find=r_addr)

        assert len(xpl.found) == 0


if __name__ == "__main__":
    unittest.main()



import imp
import os
import sys
import datetime
import gc
import ctypes

from .import_hooks import remove_fake_pkg_resources


have_gitpython = False
try:
    from git import Repo, InvalidGitRepositoryError

    have_gitpython = True
except ImportError:
    print("If you install gitpython (`pip install gitpython`), I can give you git info too!")


angr_modules = ["angr", "ailment", "cle", "pyvex", "claripy", "archinfo", "z3", "unicorn"]
native_modules = {
    "angr": lambda: angr.state_plugins.unicorn_engine._UC_NATIVE,  # pylint: disable=undefined-variable
    "unicorn": lambda: unicorn.unicorn._uc,  # pylint: disable=undefined-variable
    "pyvex": lambda: pyvex.pvc,  # pylint: disable=undefined-variable
    "z3": lambda: [x for x in gc.get_objects() if type(x) is ctypes.CDLL and "z3" in str(x)][0],  # YIKES FOREVER
}
python_packages = {"z3": "z3-solver"}


def get_venv():
    if "VIRTUAL_ENV" in os.environ:
        return os.environ["VIRTUAL_ENV"]
    return None


def print_versions():
    remove_fake_pkg_resources()
    # import the real pkg_resources
    import pkg_resources  # pylint:disable=import-outside-toplevel

    for m in angr_modules:
        print("######## %s #########" % m)
        try:
            _, python_filename, _ = imp.find_module(m)
        except ImportError:
            print("Python could not find " + m)
            continue
        except Exception as e:  # pylint: disable=broad-except
            print(f"An error occurred importing {m}: {e}")
        print("Python found it in %s" % (python_filename))
        try:
            pip_package = python_packages.get(m, m)
            pip_version = pkg_resources.get_distribution(pip_package)
            print("Pip version %s" % pip_version)
        except Exception:  # pylint: disable-broad-except
            print("Pip version not found!")
        print_git_info(python_filename)


def print_git_info(dirname):
    if not have_gitpython:
        return
    try:
        repo = Repo(dirname, search_parent_directories=True)
    except InvalidGitRepositoryError:
        print("Couldn't find git info")
        return
    cur_commit = repo.commit()
    cur_branch = repo.active_branch
    print("Git info:")
    print(f"\tCurrent commit {cur_commit.hexsha} from branch {cur_branch.name}")
    try:
        # EDG: Git is insane, but this should work 99% of the time
        cur_tb = cur_branch.tracking_branch()
        if cur_tb.is_remote():
            remote_name = cur_tb.remote_name
            remote_url = repo.remotes[remote_name].url
            print(f"\tChecked out from remote {remote_name}: {remote_url}")
        else:
            print("Tracking local branch %s" % cur_tb.name)
    except Exception:  # pylint: disable=broad-except
        print("Could not resolve tracking branch or remote info!")


def print_system_info():
    remove_fake_pkg_resources()
    # import the real pkg_resources
    import pkg_resources  # pylint:disable=import-outside-toplevel

    print("Platform: " + pkg_resources.get_build_platform())
    print("Python version: " + str(sys.version))


def print_native_info():
    print("######### Native Module Info ##########")
    for module, funcs in native_modules.items():
        try:
            globals()[module] = __import__(module)
            try:
                print(f"{module}: {funcs()}")
            except Exception as e:  # pylint: disable=broad-except
                print(f"{module}: imported but path finding raised a {type(e)}: {e}")
        except ModuleNotFoundError:
            print(f"{module}: NOT FOUND")
        except ImportError:
            print(f"{module}: FOUND BUT FAILED TO IMPORT")
        except Exception as e:  # pylint: disable=broad-except
            print(f"{module}: __import__ raised a {type(e)}: {e}")


def bug_report():
    print("angr environment report")
    print("=============================")
    print("Date: " + str(datetime.datetime.today()))
    if get_venv():
        print("Running in virtual environment at " + get_venv())
    else:
        print("!!! running in global environment.  Are you sure? !!!")
    print_system_info()
    print_versions()
    print_native_info()


if __name__ == "__main__":
    bug_report()



#!/usr/bin/env python3
__package__ = __package__ or "tests.analyses.cfg"  # pylint:disable=redefined-builtin

import time
import pickle
import networkx

import logging
import os
import unittest

import angr
from angr import options as o

from ...common import bin_location, broken

l = logging.getLogger("angr.tests.test_cfgemulated")

test_location = os.path.join(bin_location, "tests")


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestCfgemulate(unittest.TestCase):
    def compare_cfg(self, standard, g, function_list):
        """
        Standard graph comes with addresses only, and it is based on instructions, not on basic blocks
        """

        def get_function_name(addr):
            start = 0
            end = len(function_list) - 1

            while start <= end:
                mid = (start + end) / 2
                f = function_list[mid]
                if addr < f["start"]:
                    end = mid - 1
                elif addr > f["end"]:
                    start = mid + 1
                else:
                    return f["name"]

            return None

        # Sort function list
        function_list = sorted(function_list, key=lambda x: x["start"])

        # Convert the IDA-style CFG into VEX-style CFG
        s_graph = networkx.DiGraph()
        all_nodes = sorted(standard.nodes())
        addr_to_basicblock = {}
        last_basicblock = None
        for n in all_nodes:
            if last_basicblock is None:
                last_basicblock = (n, n)

            block = last_basicblock
            successors = standard.successors(n)
            if len(successors) == 1 and successors[0] >= block[0]:
                last_basicblock = (block[0], successors[0])
            else:
                # Save the existing block
                addr_to_basicblock[block[0]] = block

                # Create edges
                for s in successors:
                    s_graph.add_edge(block[0], s)

                # Clear last_basicblock so that we create a new basicblock next time
                last_basicblock = None

        graph = networkx.DiGraph()
        for src, dst in g.edges():
            graph.add_edge(src.addr, dst.addr)

        # Graph comparison
        for src, dst in s_graph.edges():
            if graph.has_edge(src, dst):
                continue
            else:
                # Edge doesn't exist in our CFG
                l.error(
                    "Edge (%s-0x%x, %s-0x%x) only exists in IDA CFG.",
                    get_function_name(src),
                    src,
                    get_function_name(dst),
                    dst,
                )

        for src, dst in graph.edges():
            if s_graph.has_edge(src, dst):
                continue
            else:
                # Edge doesn't exist in our CFG
                l.error(
                    "Edge (%s-0x%x, %s-0x%x) only exists in angr's CFG.",
                    get_function_name(src),
                    src,
                    get_function_name(dst),
                    dst,
                )

    def perform_single(self, binary_path, cfg_path=None):
        proj = angr.Project(
            binary_path,
            use_sim_procedures=True,
            default_analysis_mode="symbolic",
            load_options={"auto_load_libs": False},
        )
        start = time.time()
        cfg = proj.analyses.CFGEmulated(context_sensitivity_level=1, fail_fast=True)
        end = time.time()
        duration = end - start
        bbl_dict = cfg.nodes()

        l.info("CFG generated in %f seconds.", duration)
        l.info("Contains %d members in BBL dict.", len(bbl_dict))

        if cfg_path is not None and os.path.isfile(cfg_path):
            # Compare the graph with a predefined CFG
            info = pickle.load(open(cfg_path, "rb"))
            standard = info["cfg"]
            functions = info["functions"]
            graph = cfg.graph

            self.compare_cfg(standard, graph, functions)
        else:
            l.warning("No standard CFG specified.")

    @broken
    def test_cfg_0(self):
        binary_path = os.path.join(test_location, "x86_64", "cfg_0")
        cfg_path = binary_path + ".cfg"
        self.perform_single(binary_path, cfg_path)

    @broken
    def test_cfg_1(self):
        binary_path = os.path.join(test_location, "x86_64", "cfg_1")
        cfg_path = binary_path + ".cfg"
        self.perform_single(binary_path, cfg_path)

    @broken
    def test_cfg_2(self):
        binary_path = os.path.join(test_location, "armel", "test_division")
        cfg_path = binary_path + ".cfg"
        self.perform_single(binary_path, cfg_path)

    @broken
    def test_cfg_3(self):
        binary_path = os.path.join(test_location, "mips", "test_arrays")
        cfg_path = binary_path + ".cfg"
        self.perform_single(binary_path, cfg_path)

    @broken
    def test_cfg_4(self):
        binary_path = os.path.join(test_location, "mipsel", "darpa_ping")
        cfg_path = binary_path + ".cfg"
        self.perform_single(binary_path, cfg_path)

    def test_additional_edges(self):
        # Test the `additional_edges` parameter for CFG generation

        binary_path = os.path.join(test_location, "x86_64", "switch")
        proj = angr.Project(
            binary_path,
            use_sim_procedures=True,
            default_analysis_mode="symbolic",
            load_options={"auto_load_libs": False},
        )
        additional_edges = {0x400573: [0x400580, 0x40058F, 0x40059E]}
        cfg = proj.analyses.CFGEmulated(
            context_sensitivity_level=0,
            additional_edges=additional_edges,
            fail_fast=True,
            resolve_indirect_jumps=False,  # For this test case, we need to disable the
            # jump table resolving, otherwise CFGEmulated
            # can automatically find the node 0x4005ad.
        )

        assert cfg.get_any_node(0x400580) is not None
        assert cfg.get_any_node(0x40058F) is not None
        assert cfg.get_any_node(0x40059E) is not None
        assert cfg.get_any_node(0x4005AD) is None

    def test_not_returning(self):
        # Make sure we are properly labeling functions that do not return in function manager

        binary_path = os.path.join(test_location, "x86_64", "not_returning")
        proj = angr.Project(binary_path, use_sim_procedures=True, load_options={"auto_load_libs": False})
        proj.analyses.CFGEmulated(context_sensitivity_level=0, fail_fast=True)  # pylint:disable=unused-variable

        # function_a returns
        assert proj.kb.functions.function(name="function_a") is not None
        assert proj.kb.functions.function(name="function_a").returning

        # function_b does not return
        assert proj.kb.functions.function(name="function_b") is not None
        assert not proj.kb.functions.function(name="function_b").returning

        # function_c does not return
        assert proj.kb.functions.function(name="function_c") is not None
        assert not proj.kb.functions.function(name="function_c").returning

        # main does not return
        assert proj.kb.functions.function(name="main") is not None
        assert not proj.kb.functions.function(name="main").returning

        # function_d should not be reachable
        assert proj.kb.functions.function(name="function_d") is None

    @broken
    def test_cfg_5(self):
        binary_path = os.path.join(test_location, "mipsel", "busybox")
        cfg_path = binary_path + ".cfg"

        self.perform_single(binary_path, cfg_path)

    def test_cfg_6(self):
        function_addresses = [
            0xFA630,
            0xFA683,
            0xFA6D4,
            0xFA707,
            0xFA754,
            0xFA779,
            0xFA7A9,
            0xFA7D6,
            0xFA844,
            0xFA857,
            0xFA8D9,
            0xFA92F,
            0xFA959,
            0xFA9FB,
            0xFABD6,
            0xFAC61,
            0xFACC2,
            0xFAD29,
            0xFAF94,
            0xFBD07,
            0xFC100,
            0xFC101,
            0xFC14F,
            0xFC18E,
            0xFC25E,
            0xFC261,
            0xFC3C6,
            0xFC42F,
            0xFC4A3,
            0xFC4CF,
            0xFC4DB,
            0xFC5BA,
            0xFC5EF,
            0xFC5FE,
            0xFC611,
            0xFC682,
            0xFC6B7,
            0xFC7FC,
            0xFC8A8,
            0xFC8E7,
            0xFCB42,
            0xFCB50,
            0xFCB72,
            0xFCC3B,
            0xFCC7A,
            0xFCC8B,
            0xFCCDC,
            0xFD1A3,
            0xFF06E,
        ]

        # We need to add DO_CCALLS to resolve long jmp and support real mode
        o.modes["fastpath"] |= {o.DO_CCALLS}
        binary_path = test_location + "/i386/bios.bin.elf"
        proj = angr.Project(binary_path, use_sim_procedures=True, page_size=1, auto_load_libs=False)
        proj.analyses.CFGEmulated(context_sensitivity_level=1, fail_fast=True)  # pylint:disable=unused-variable
        assert {f for f in proj.kb.functions} >= set(function_addresses)
        o.modes["fastpath"] ^= {o.DO_CCALLS}

    def test_fauxware(self):
        binary_path = os.path.join(test_location, "x86_64", "fauxware")
        cfg_path = binary_path + ".cfg"

        self.perform_single(binary_path, cfg_path)

    @broken
    def test_loop_unrolling(self):
        binary_path = os.path.join(test_location, "x86_64", "cfg_loop_unrolling")

        p = angr.Project(binary_path, auto_load_libs=True)
        cfg = p.analyses.CFGEmulated(fail_fast=True)

        cfg.normalize()
        cfg.unroll_loops(5)

        assert len(cfg.get_all_nodes(0x400636)) == 7

    def test_thumb_mode(self):
        # In thumb mode, all addresses of instructions and in function manager should be odd numbers, which loyally
        # reflect VEX's trick to encode the THUMB state in the address.

        binary_path = os.path.join(test_location, "armhf", "test_arrays")
        p = angr.Project(binary_path, auto_load_libs=False)
        cfg = p.analyses.CFGEmulated(fail_fast=True)

        def check_addr(a):
            if a % 2 == 1:
                assert cfg.is_thumb_addr(a)
            else:
                assert not cfg.is_thumb_addr(a)

        # CFGNodes
        cfg_node_addrs = [n.addr for n in cfg.graph.nodes() if not n.is_simprocedure]
        for a in cfg_node_addrs:
            check_addr(a)

        # Functions in function manager
        for f_addr, f in p.kb.functions.items():
            if f.is_simprocedure:
                continue
            check_addr(f_addr)
            if f.startpoint is not None:
                check_addr(f.startpoint.addr)

    def test_fakeret_edges_0(self):
        # Test the bug where a fakeret edge can be missing in certain cases
        # Reported by Attila Axt (GitHub: @axt)
        # Ref: https://github.com/angr/angr/issues/72

        binary_path = os.path.join(test_location, "x86_64", "cfg_3")

        p = angr.Project(binary_path, auto_load_libs=False)
        cfg = p.analyses.CFGEmulated(context_sensitivity_level=3, fail_fast=True)

        putchar_plt = cfg.functions.function(name="putchar", plt=True)
        assert putchar_plt.returning

        putchar = cfg.functions.function(name="putchar", plt=False)
        assert putchar.returning

        # Since context sensitivity is 3, there should be two different putchar nodes
        putchar_cfgnodes = cfg.get_all_nodes(putchar.addr)
        assert len(putchar_cfgnodes) == 2

        # Each putchar node has a different predecessor as their PLT entry
        plt_entry_0 = cfg.get_predecessors(putchar_cfgnodes[0])
        assert len(plt_entry_0) == 1
        plt_entry_0 = plt_entry_0[0]

        plt_entry_1 = cfg.get_predecessors(putchar_cfgnodes[1])
        assert len(plt_entry_1) == 1
        plt_entry_1 = plt_entry_1[0]

        assert plt_entry_0 is not plt_entry_1

        # Each PLT entry should have a FakeRet edge
        preds_0 = cfg.get_predecessors(plt_entry_0)
        assert len(preds_0) == 1
        preds_1 = cfg.get_predecessors(plt_entry_1)
        assert len(preds_1) == 1

        # Each predecessor must have a call edge and a FakeRet edge
        edges_0 = cfg.get_successors_and_jumpkind(preds_0[0], excluding_fakeret=False)
        assert len(edges_0) == 2
        jumpkinds = {jumpkind for _, jumpkind in edges_0}
        assert jumpkinds == {"Ijk_Call", "Ijk_FakeRet"}

        edges_1 = cfg.get_successors_and_jumpkind(preds_1[0], excluding_fakeret=False)
        assert len(edges_1) == 2
        jumpkinds = {jumpkind for _, jumpkind in edges_1}
        assert jumpkinds == {"Ijk_Call", "Ijk_FakeRet"}

    def test_string_references(self):
        # Test AttributeError on 'addr' which occurs when searching for string
        # references

        binary_path = os.path.join(test_location, "i386", "ctf_nuclear")
        b = angr.Project(binary_path, load_options={"auto_load_libs": False})
        cfg = b.analyses.CFGEmulated(keep_state=True, fail_fast=True)

        string_references = []
        for f in cfg.functions.values():
            string_references.append(f.string_references())

        # test passes if hasn't thrown an exception

    def test_arrays(self):
        binary_path = os.path.join(test_location, "armhf", "test_arrays")
        b = angr.Project(binary_path, load_options={"auto_load_libs": False})
        cfg = b.analyses.CFGEmulated(fail_fast=True)

        node = cfg.model.get_any_node(0x10415)
        assert node is not None

        successors = cfg.model.get_successors(node)
        assert len(successors) == 2

    def test_max_steps(self):
        binary_path = os.path.join(test_location, "x86_64", "fauxware")
        b = angr.Project(binary_path, load_options={"auto_load_libs": False})
        cfg = b.analyses.CFGEmulated(max_steps=5, fail_fast=True)

        dfs_edges = networkx.dfs_edges(cfg.graph)

        depth_map = {}
        for src, dst in dfs_edges:
            if src not in depth_map:
                depth_map[src] = 0
            if dst not in depth_map:
                depth_map[dst] = depth_map[src] + 1
            depth_map[dst] = max(depth_map[src] + 1, depth_map[dst])

        assert max(depth_map.values()) <= 5

    def test_armel_final_missing_block(self):
        # Due to a stupid bug in CFGEmulated, the last block of a function might go missing in the function graph if the
        # only entry edge to that block is an Ijk_Ret edge. See #475 on GitHub.
        # Thank @gergo for reporting and providing this test binary.

        binary_path = os.path.join(test_location, "armel", "last_block")
        b = angr.Project(binary_path, auto_load_libs=False)
        cfg = b.analyses.CFGEmulated(fail_fast=True)

        blocks = list(cfg.kb.functions[0x8000].blocks)

        assert len(blocks) == 3
        assert {block.addr for block in blocks} == {0x8000, 0x8014, 0x8020}

    def test_armel_final_missing_block_b(self):
        # When _pending_jobs is not sorted, it is possible that we first process a pending job created earlier and then
        # process another pending job created later. Ideally, we hope that jobs are always processed in a topological
        # order, and the unsorted pending jobs break this assumption. In this test binary, at one point there can be two
        # pending jobs, 0x10b05/0x10ac5(Ijk_FakeRet) and 0x10bbe(Ijk_FakeRet). If 0x10bbe is processed before 0x10b05,
        # we do not # know whether the function 0x10a29(aes) returns or not. As a result, the final block of the main
        # function is not confirmed, and is not added to the function graph of function main.
        #
        # In fact, this also hints a different bug. We should always "confirm" that a function returns if its FakeRet
        # job are processed for whatever reason.
        #
        # Fixing either bug will resolve the issue that the final block does not show up in the function graph of main.
        # To stay on the safe side, both of them are fixed. Thanks @tyb0807 for reporting this issue and providing a
        # test binary.
        # EDG says: This binary is compiled incorrectly.
        # The binary's app code was compiled as CortexM, but linked against ARM libraries.
        # This is illegal, and does not actually execute on a real CortexM.
        # Somebody should recompile it....
        binary_path = os.path.join(test_location, "armel", "aes")
        b = angr.Project(binary_path, arch="ARMEL", auto_load_libs=False)

        function = b.loader.main_object.get_symbol("main").rebased_addr
        cfg = b.analyses.CFGEmulated(
            starts=[function],
            context_sensitivity_level=0,
            normalize=True,
            fail_fast=True,
        )

        blocks = list(cfg.kb.functions["main"].blocks)

        assert len(blocks) == 2
        assert {block.addr for block in blocks} == {0x10B79, 0x10BBF}

    def test_armel_incorrect_function_detection_caused_by_branch(self):
        # GitHub issue #685
        binary_path = os.path.join(test_location, "armel", "RTOSDemo.axf.issue_685")
        b = angr.Project(binary_path, auto_load_libs=False)

        cfg = b.analyses.CFGEmulated()

        # The Main function should be identified as a single function
        assert 0x80A1 in cfg.functions
        main_func = cfg.functions[0x80A1]

        # All blocks should be there
        block_addrs = sorted([b.addr for b in main_func.blocks])
        assert block_addrs == [0x80A1, 0x80B1, 0x80BB, 0x80CD, 0x80DF, 0x80E3, 0x80ED]

        # The ResetISR function should be identified as a single function, too
        assert 0x8009 in cfg.functions
        resetisr_func = cfg.functions[0x8009]

        # All blocks should be there
        block_addrs = sorted([b.addr for b in resetisr_func.blocks])
        assert block_addrs == [0x8009, 0x8011, 0x801F, 0x8027]

    def test_cfg_switches(self):
        # logging.getLogger('angr.analyses.cfg.cfg_fast').setLevel(logging.INFO)
        # logging.getLogger('angr.analyses.cfg.indirect_jump_resolvers.jumptable').setLevel(logging.DEBUG)

        filename = "cfg_switches"

        edges = {
            "x86_64": {
                # jump table 0 in func_0
                (0x40053A, 0x400547),
                (0x40053A, 0x400552),
                (0x40053A, 0x40055D),
                (0x40053A, 0x400568),
                (0x40053A, 0x400573),
                (0x40053A, 0x400580),
                (0x40053A, 0x40058D),
                # jump table 0 in func_1
                (0x4005BC, 0x4005C9),
                (0x4005BC, 0x4005D8),
                (0x4005BC, 0x4005E7),
                (0x4005BC, 0x4005F6),
                (0x4005BC, 0x400605),
                (0x4005BC, 0x400614),
                (0x4005BC, 0x400623),
                (0x4005BC, 0x400632),
                (0x4005BC, 0x40063E),
                (0x4005BC, 0x40064A),
                (0x4005BC, 0x4006B0),
                # jump table 1 in func_1
                (0x40065A, 0x400667),
                (0x40065A, 0x400673),
                (0x40065A, 0x40067F),
                (0x40065A, 0x40068B),
                (0x40065A, 0x400697),
                (0x40065A, 0x4006A3),
                # jump table 0 in main
                (0x4006E1, 0x4006EE),
                (0x4006E1, 0x4006FA),
                (0x4006E1, 0x40070B),
                (0x4006E1, 0x40071C),
                (0x4006E1, 0x40072D),
                (0x4006E1, 0x40073E),
                (0x4006E1, 0x40074F),
                (0x4006E1, 0x40075B),
            },
        }

        arches = edges.keys()

        for arch in arches:
            path = os.path.join(test_location, arch, filename)
            proj = angr.Project(path, load_options={"auto_load_libs": False})

            cfg = proj.analyses.CFGEmulated()

            for src, dst in edges[arch]:
                src_node = cfg.get_any_node(src)
                dst_node = cfg.get_any_node(dst)
                assert dst_node in src_node.successors, "CFG edge {}-{} is not found.".format(
                    src_node,
                    dst_node,
                )

    class CFGEmulatedAborted(angr.analyses.cfg.cfg_emulated.CFGEmulated):  # pylint:disable=abstract-method
        """
        Only used in the test_abort_and_resume test case.
        """

        should_abort = False

        def _intra_analysis(self):
            if self.should_abort:
                self.abort()
            else:
                super()._intra_analysis()

    def test_abort_and_resume(self):
        angr.analyses.AnalysesHub.register_default("CFGEmulatedAborted", self.CFGEmulatedAborted)

        self.CFGEmulatedAborted.should_abort = False
        binary_path = os.path.join(test_location, "x86_64", "fauxware")
        b = angr.Project(binary_path, auto_load_libs=False)

        self.CFGEmulatedAborted.should_abort = True
        cfg = b.analyses.CFGEmulatedAborted()
        assert len(list(cfg.jobs)) > 0  # there should be left-over jobs

        self.CFGEmulatedAborted.should_abort = False
        cfg.resume()

        assert len(list(cfg.jobs)) == 0

    def test_base_graph(self):
        path = os.path.join(test_location, "x86_64", "test_cfgemulated_base_graph")

        func_addr = 0x401129

        edges = {
            (0x401129, 0x401144),
            (0x401129, 0x40114D),
            (0x401144, 0x401154),
            (0x40114D, 0x401154),
        }

        final_states_info = {
            0x401129: 2,
            0x40114D: 1,
            0x401144: 1,
            0x401154: 1,
        }

        proj = angr.Project(path, load_options={"auto_load_libs": False})

        cfg_fast = proj.analyses.CFGFast(normalize=True)
        target_function = cfg_fast.kb.functions[func_addr]
        target_function.normalize()

        target_function_cfg_emulated = proj.analyses.CFGEmulated(
            keep_state=True,
            state_add_options=angr.options.refs,
            base_graph=target_function.graph,
            starts=(func_addr,),
            normalize=True,
        )
        for src, dst in edges:
            src_node = target_function_cfg_emulated.get_any_node(src)
            dst_node = target_function_cfg_emulated.get_any_node(dst)
            assert dst_node in src_node.successors, "CFG edge {}-{} is not found.".format(
                src_node,
                dst_node,
            )

        for node_addr, final_states_number in final_states_info.items():
            node = target_function_cfg_emulated.get_any_node(node_addr)
            assert final_states_number == len(node.final_states), (
                "CFG node 0x%x has incorrect final states." % node_addr
            )


if __name__ == "__main__":
    logging.getLogger("angr.state_plugins.abstract_memory").setLevel(logging.DEBUG)
    # logging.getLogger("angr.state_plugins.symbolic_memory").setLevel(logging.DEBUG)
    # logging.getLogger("angr.analyses.cfg.cfg_emulated").setLevel(logging.DEBUG)
    # logging.getLogger("s_irsb").setLevel(logging.DEBUG)
    # Temporarily disable the warnings of claripy backend
    # logging.getLogger("claripy.backends.backend").setLevel(logging.ERROR)
    # logging.getLogger("claripy.claripy").setLevel(logging.ERROR)

    unittest.main()



#!/usr/bin/env python3
__package__ = __package__ or "tests.sim.options"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class Test0Div(unittest.TestCase):
    def _run_0div(self, arch):
        # check that we run in unicorn up to the zero-div site, fall back, try again in angr, and error correctly.
        p = angr.Project(os.path.join(test_location, arch, "test_0div"), auto_load_libs=False)
        s = p.factory.entry_state(add_options=angr.options.unicorn)
        simgr = p.factory.simulation_manager(s)
        simgr.run(n=5)
        assert len(simgr.active) == 1
        simgr.step()
        assert len(simgr.errored) == 1
        assert isinstance(simgr.errored[0].error, angr.errors.SimZeroDivisionException)

    def test_0div_i386(self):
        self._run_0div("i386")

    def test_0div_x86_64(self):
        self._run_0div("x86_64")

    def test_symbolic_0div(self):
        p = angr.load_shellcode(b"X", arch="amd64")
        s = p.factory.blank_state()
        s.regs.rax = s.solver.BVS("rax", 64)
        s.regs.rcx = s.solver.BVS("rcx", 64)
        s.regs.rdx = s.solver.BVS("rdx", 64)

        s.options.add(angr.options.PRODUCE_ZERODIV_SUCCESSORS)
        successors = s.step(insn_bytes=b"\x48\xf7\xf1")  # div rcx
        assert len(successors.flat_successors) == 2

        s.options.discard(angr.options.PRODUCE_ZERODIV_SUCCESSORS)
        successors = s.step(insn_bytes=b"\x48\xf7\xf1")  # div rcx
        assert len(successors.flat_successors) == 1


if __name__ == "__main__":
    unittest.main()



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("tokenbinding.dll")
prototypes = \
    {
        #
        'TokenBindingGenerateBinding': SimTypeFunction([SimTypeInt(signed=False, label="TOKENBINDING_KEY_PARAMETERS_TYPE"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="TOKENBINDING_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="TOKENBINDING_EXTENSION_FORMAT"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"bindingType": SimTypeInt(signed=False, label="TOKENBINDING_TYPE"), "identifierSize": SimTypeInt(signed=False, label="UInt32"), "identifierData": SimTypePointer(SimStruct({"keyType": SimTypeChar(label="Byte")}, name="TOKENBINDING_IDENTIFIER", pack=False, align=None), offset=0), "extensionFormat": SimTypeInt(signed=False, label="TOKENBINDING_EXTENSION_FORMAT"), "extensionSize": SimTypeInt(signed=False, label="UInt32"), "extensionData": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="TOKENBINDING_RESULT_DATA", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["keyType", "targetURL", "bindingType", "tlsEKM", "tlsEKMSize", "extensionFormat", "extensionData", "tokenBinding", "tokenBindingSize", "resultData"]),
        #
        'TokenBindingGenerateMessage': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["tokenBindings", "tokenBindingsSize", "tokenBindingsCount", "tokenBindingMessage", "tokenBindingMessageSize"]),
        #
        'TokenBindingVerifyMessage': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="TOKENBINDING_KEY_PARAMETERS_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"resultCount": SimTypeInt(signed=False, label="UInt32"), "resultData": SimTypePointer(SimStruct({"bindingType": SimTypeInt(signed=False, label="TOKENBINDING_TYPE"), "identifierSize": SimTypeInt(signed=False, label="UInt32"), "identifierData": SimTypePointer(SimStruct({"keyType": SimTypeChar(label="Byte")}, name="TOKENBINDING_IDENTIFIER", pack=False, align=None), offset=0), "extensionFormat": SimTypeInt(signed=False, label="TOKENBINDING_EXTENSION_FORMAT"), "extensionSize": SimTypeInt(signed=False, label="UInt32"), "extensionData": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="TOKENBINDING_RESULT_DATA", pack=False, align=None), offset=0)}, name="TOKENBINDING_RESULT_LIST", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["tokenBindingMessage", "tokenBindingMessageSize", "keyType", "tlsEKM", "tlsEKMSize", "resultList"]),
        #
        'TokenBindingGetKeyTypesClient': SimTypeFunction([SimTypePointer(SimTypePointer(SimStruct({"keyCount": SimTypeInt(signed=False, label="UInt32"), "keyType": SimTypePointer(SimTypeInt(signed=False, label="TOKENBINDING_KEY_PARAMETERS_TYPE"), offset=0)}, name="TOKENBINDING_KEY_TYPES", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["keyTypes"]),
        #
        'TokenBindingGetKeyTypesServer': SimTypeFunction([SimTypePointer(SimTypePointer(SimStruct({"keyCount": SimTypeInt(signed=False, label="UInt32"), "keyType": SimTypePointer(SimTypeInt(signed=False, label="TOKENBINDING_KEY_PARAMETERS_TYPE"), offset=0)}, name="TOKENBINDING_KEY_TYPES", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["keyTypes"]),
        #
        'TokenBindingDeleteBinding': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["targetURL"]),
        #
        'TokenBindingDeleteAllBindings': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'TokenBindingGenerateID': SimTypeFunction([SimTypeInt(signed=False, label="TOKENBINDING_KEY_PARAMETERS_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"bindingType": SimTypeInt(signed=False, label="TOKENBINDING_TYPE"), "identifierSize": SimTypeInt(signed=False, label="UInt32"), "identifierData": SimTypePointer(SimStruct({"keyType": SimTypeChar(label="Byte")}, name="TOKENBINDING_IDENTIFIER", pack=False, align=None), offset=0), "extensionFormat": SimTypeInt(signed=False, label="TOKENBINDING_EXTENSION_FORMAT"), "extensionSize": SimTypeInt(signed=False, label="UInt32"), "extensionData": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="TOKENBINDING_RESULT_DATA", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["keyType", "publicKey", "publicKeySize", "resultData"]),
        #
        'TokenBindingGenerateIDForUri': SimTypeFunction([SimTypeInt(signed=False, label="TOKENBINDING_KEY_PARAMETERS_TYPE"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"bindingType": SimTypeInt(signed=False, label="TOKENBINDING_TYPE"), "identifierSize": SimTypeInt(signed=False, label="UInt32"), "identifierData": SimTypePointer(SimStruct({"keyType": SimTypeChar(label="Byte")}, name="TOKENBINDING_IDENTIFIER", pack=False, align=None), offset=0), "extensionFormat": SimTypeInt(signed=False, label="TOKENBINDING_EXTENSION_FORMAT"), "extensionSize": SimTypeInt(signed=False, label="UInt32"), "extensionData": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="TOKENBINDING_RESULT_DATA", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["keyType", "targetUri", "resultData"]),
        #
        'TokenBindingGetHighestSupportedVersion': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["majorVersion", "minorVersion"]),
    }

lib.set_prototypes(prototypes)



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,disable=no-self-use
__package__ = __package__ or "tests.analyses"  # pylint:disable=redefined-builtin

import os.path
import unittest

import angr

from ..common import bin_location


class TestFlirt(unittest.TestCase):
    def test_amd64_elf_static_libc_ubuntu_2004(self):
        binary_path = os.path.join(bin_location, "tests", "x86_64", "elf_with_static_libc_ubuntu_2004_stripped")
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=False)
        cfg = proj.analyses.CFGFast(show_progressbar=False)  # , detect_tail_calls=True)
        flirt_path = os.path.join(bin_location, "tests", "x86_64", "libc_ubuntu_2004.sig")
        proj.analyses.Flirt(flirt_path)

        assert cfg.functions[0x415CC0].name == "_IO_file_open"
        assert cfg.functions[0x415CC0].is_default_name is False
        assert cfg.functions[0x415CC0].from_signature == "flirt"
        assert cfg.functions[0x436980].name == "__mempcpy_chk_avx512_no_vzeroupper"
        assert cfg.functions[0x436980].is_default_name is False
        assert cfg.functions[0x436980].from_signature == "flirt"

    def test_armhf_elf_static_using_armel_libc(self):
        binary_path = os.path.join(bin_location, "tests", "armhf", "amp_challenge_07.gcc")
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=False)
        proj.analyses.CFGFast(show_progressbar=False)
        flirt_path = os.path.join(bin_location, "tests", "armhf", "debian_10.3_libc.sig")
        flirt = proj.analyses.Flirt(flirt_path)

        assert len(flirt.matched_suggestions) == 1

        assert proj.kb.functions[0x1004C9].name == "strstr"
        assert proj.kb.functions[0x1004C9].prototype is not None
        assert proj.kb.functions[0x1004C9].calling_convention is not None

        assert proj.kb.functions[0xF38D9].name == "__printf"
        assert proj.kb.functions[0xF38D9].prototype is not None
        assert proj.kb.functions[0xF38D9].calling_convention is not None


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint:disable=no-self-use
__package__ = __package__ or "tests.analyses.cfg"  # pylint:disable=redefined-builtin

import os
import unittest
import logging

import angr
from angr.analyses import CFGFast

from ...common import bin_location

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
FAUXWARE_PATH = os.path.join(bin_location, "tests", "x86_64", "fauxware")


class TestCfgModel(unittest.TestCase):
    """
    Test cases for CFGModel.
    """

    def test_cfgmodel_clear_region_for_reflow(self):
        """Test CFGModel::clear_region_for_reflow."""
        proj = angr.Project(FAUXWARE_PATH, auto_load_libs=False)
        cfg = proj.analyses[CFGFast].prep()()
        func = cfg.functions["authenticate"]

        addr = 0x40068E
        end_addr = 0x400692
        cfg.model.clear_region_for_reflow(addr, end_addr - addr)

        for addr in func.block_addrs:
            assert cfg.model.get_any_node(addr) is None

    def test_cfgmodel_clear_region_for_reflow_multifunc(self):
        """Test CFGModel::clear_region_for_reflow across function boundaries."""
        proj = angr.Project(FAUXWARE_PATH, auto_load_libs=False)
        cfg = proj.analyses[CFGFast].prep()()

        expected_removed_addrs = list(cfg.functions["accepted"].block_addrs) + list(
            cfg.functions["rejected"].block_addrs
        )

        addr = 0x4006FB
        end_addr = 0x4006FE
        cfg.model.clear_region_for_reflow(addr, end_addr - addr)

        for addr in expected_removed_addrs:
            assert cfg.model.get_any_node(addr) is None


if __name__ == "__main__":
    unittest.main()



default_plugins = {}


class KnowledgeBasePlugin:
    def copy(self):
        raise NotImplementedError

    @staticmethod
    def register_default(name, cls):
        if name in default_plugins:
            raise Exception(f"{default_plugins[name]} is already set as the default for {name}")
        default_plugins[name] = cls
